# -*- coding: utf-8 -*-
"""全球货币枚举"""
from collections import namedtuple
from enum import Enum, auto
from typing import Dict, List
from .base import ByKey

# https://www.iso.org/iso-4217-currency-codes.html

field_names: List[str] = ["alphabetic_code", "en_name", "numeric_code", "minor_unit"]
CurrencyDetail = namedtuple(typename="CurrencyDetail", field_names=field_names)
tsv: str = """
AED\tUAE Dirham\t784\t2
AFN\tAfghani\t971\t2
ALL\tLek\t008\t2
AMD\tArmenian Dram\t051\t2
ANG\tNetherlands Antillean Guilder\t532\t2
AOA\tKwanza\t973\t2
ARS\tArgentine Peso\t032\t2
AUD\tAustralian Dollar\t036\t2
AWG\tAruban Florin\t533\t2
AZN\tAzerbaijan Manat\t944\t2
BAM\tConvertible Mark\t977\t2
BBD\tBarbados Dollar\t052\t2
BDT\tTaka\t050\t2
BGN\tBulgarian Lev\t975\t2
BHD\tBahraini Dinar\t048\t3
BIF\tBurundi Franc\t108\t0
BMD\tBermudian Dollar\t060\t2
BND\tBrunei Dollar\t096\t2
BOB\tBoliviano\t068\t2
BRL\tBrazilian Real\t986\t2
BSD\tBahamian Dollar\t044\t2
BTN\tNgultrum\t064\t2
BWP\tPula\t072\t2
BYN\tBelarusian Ruble\t933\t2
BZD\tBelize Dollar\t084\t2
CAD\tCanadian Dollar\t124\t2
CDF\tCongolese Franc\t976\t2
CHF\tSwiss Franc\t756\t2
CLP\tChilean Peso\t152\t0
CNY\tYuan Renminbi\t156\t2
COP\tColombian Peso\t170\t2
CRC\tCosta Rican Colon\t188\t2
CUC\tPeso Convertible\t931\t2
CUP\tCuban Peso\t192\t2
CVE\tCabo Verde Escudo\t132\t2
CZK\tCzech Koruna\t203\t2
DJF\tDjibouti Franc\t262\t0
DKK\tDanish Krone\t208\t2
DOP\tDominican Peso\t214\t2
DZD\tAlgerian Dinar\t012\t2
EGP\tEgyptian Pound\t818\t2
ERN\tNakfa\t232\t2
ETB\tEthiopian Birr\t230\t2
EUR\tEuro\t978\t2
FJD\tFiji Dollar\t242\t2
FKP\tFalkland Islands Pound\t238\t2
GBP\tPound Sterling\t826\t2
GEL\tLari\t981\t2
GHS\tGhana Cedi\t936\t2
GIP\tGibraltar Pound\t292\t2
GMD\tDalasi\t270\t2
GNF\tGuinean Franc\t324\t0
GTQ\tQuetzal\t320\t2
GYD\tGuyana Dollar\t328\t2
HKD\tHong Kong Dollar\t344\t2
HNL\tLempira\t340\t2
HTG\tGourde\t332\t2
HUF\tForint\t348\t2
IDR\tRupiah\t360\t2
ILS\tNew Israeli Sheqel\t376\t2
INR\tIndian Rupee\t356\t2
IQD\tIraqi Dinar\t368\t3
IRR\tIranian Rial\t364\t2
ISK\tIceland Krona\t352\t0
JMD\tJamaican Dollar\t388\t2
JOD\tJordanian Dinar\t400\t3
JPY\tYen\t392\t0
KES\tKenyan Shilling\t404\t2
KGS\tSom\t417\t2
KHR\tRiel\t116\t2
KMF\tComorian Franc \t174\t0
KPW\tNorth Korean Won\t408\t2
KRW\tWon\t410\t0
KWD\tKuwaiti Dinar\t414\t3
KYD\tCayman Islands Dollar\t136\t2
KZT\tTenge\t398\t2
LAK\tLao Kip\t418\t2
LBP\tLebanese Pound\t422\t2
LKR\tSri Lanka Rupee\t144\t2
LRD\tLiberian Dollar\t430\t2
LSL\tLoti\t426\t2
LYD\tLibyan Dinar\t434\t3
MAD\tMoroccan Dirham\t504\t2
MDL\tMoldovan Leu\t498\t2
MGA\tMalagasy Ariary\t969\t2
MKD\tDenar\t807\t2
MMK\tKyat\t104\t2
MNT\tTugrik\t496\t2
MOP\tPataca\t446\t2
MRU\tOuguiya\t929\t2
MUR\tMauritius Rupee\t480\t2
MVR\tRufiyaa\t462\t2
MWK\tMalawi Kwacha\t454\t2
MXN\tMexican Peso\t484\t2
MYR\tMalaysian Ringgit\t458\t2
MZN\tMozambique Metical\t943\t2
NAD\tNamibia Dollar\t516\t2
NGN\tNaira\t566\t2
NIO\tCordoba Oro\t558\t2
NOK\tNorwegian Krone\t578\t2
NPR\tNepalese Rupee\t524\t2
NZD\tNew Zealand Dollar\t554\t2
OMR\tRial Omani\t512\t3
PAB\tBalboa\t590\t2
PEN\tSol\t604\t2
PGK\tKina\t598\t2
PHP\tPhilippine Peso\t608\t2
PKR\tPakistan Rupee\t586\t2
PLN\tZloty\t985\t2
PYG\tGuarani\t600\t0
QAR\tQatari Rial\t634\t2
RON\tRomanian Leu\t946\t2
RSD\tSerbian Dinar\t941\t2
RUB\tRussian Ruble\t643\t2
RWF\tRwanda Franc\t646\t0
SAR\tSaudi Riyal\t682\t2
SBD\tSolomon Islands Dollar\t090\t2
SCR\tSeychelles Rupee\t690\t2
SDG\tSudanese Pound\t938\t2
SEK\tSwedish Krona\t752\t2
SGD\tSingapore Dollar\t702\t2
SHP\tSaint Helena Pound\t654\t2
SLE\tLeone\t925\t2
SLL\tLeone\t694\t2
SOS\tSomali Shilling\t706\t2
SRD\tSurinam Dollar\t968\t2
SSP\tSouth Sudanese Pound\t728\t2
STN\tDobra\t930\t2
SVC\tEl Salvador Colon\t222\t2
SYP\tSyrian Pound\t760\t2
SZL\tLilangeni\t748\t2
THB\tBaht\t764\t2
TJS\tSomoni\t972\t2
TMT\tTurkmenistan New Manat\t934\t2
TND\tTunisian Dinar\t788\t3
TOP\tPa’anga\t776\t2
TRY\tTurkish Lira\t949\t2
TTD\tTrinidad and Tobago Dollar\t780\t2
TWD\tNew Taiwan Dollar\t901\t2
TZS\tTanzanian Shilling\t834\t2
UAH\tHryvnia\t980\t2
UGX\tUganda Shilling\t800\t0
USD\tUS Dollar\t840\t2
UYU\tPeso Uruguayo\t858\t2
UYW\tUnidad Previsional\t927\t4
UZS\tUzbekistan Sum\t860\t2
VED\tBolívar Soberano\t926\t2
VES\tBolívar Soberano\t928\t2
VND\tDong\t704\t0
VUV\tVatu\t548\t0
WST\tTala\t882\t2
XAF\tCFA Franc BEAC\t950\t0
XCD\tEast Caribbean Dollar\t951\t2
XOF\tCFA Franc BCEAO\t952\t0
XPF\tCFP Franc\t953\t0
YER\tYemeni Rial\t886\t2
ZAR\tRand\t710\t2
ZMW\tZambian Kwacha\t967\t2
ZWL\tZimbabwe Dollar\t932\t2
"""

Items: List[CurrencyDetail] = [CurrencyDetail(*line.strip().split("\t")) for line in tsv.splitlines() if line]


class ByAlphabeticCode(ByKey):
    map: Dict[str, CurrencyDetail] = {getattr(item, field_names[0]): item for item in Items}


class Currency(Enum):
    """交易所MIC枚举"""
    AED = ByAlphabeticCode.map['AED']
    AFN = ByAlphabeticCode.map['AFN']
    ALL = ByAlphabeticCode.map['ALL']
    AMD = ByAlphabeticCode.map['AMD']
    ANG = ByAlphabeticCode.map['ANG']
    AOA = ByAlphabeticCode.map['AOA']
    ARS = ByAlphabeticCode.map['ARS']
    AUD = ByAlphabeticCode.map['AUD']
    AWG = ByAlphabeticCode.map['AWG']
    AZN = ByAlphabeticCode.map['AZN']
    BAM = ByAlphabeticCode.map['BAM']
    BBD = ByAlphabeticCode.map['BBD']
    BDT = ByAlphabeticCode.map['BDT']
    BGN = ByAlphabeticCode.map['BGN']
    BHD = ByAlphabeticCode.map['BHD']
    BIF = ByAlphabeticCode.map['BIF']
    BMD = ByAlphabeticCode.map['BMD']
    BND = ByAlphabeticCode.map['BND']
    BOB = ByAlphabeticCode.map['BOB']
    BRL = ByAlphabeticCode.map['BRL']
    BSD = ByAlphabeticCode.map['BSD']
    BTN = ByAlphabeticCode.map['BTN']
    BWP = ByAlphabeticCode.map['BWP']
    BYN = ByAlphabeticCode.map['BYN']
    BZD = ByAlphabeticCode.map['BZD']
    CAD = ByAlphabeticCode.map['CAD']
    CDF = ByAlphabeticCode.map['CDF']
    CHF = ByAlphabeticCode.map['CHF']
    CLP = ByAlphabeticCode.map['CLP']
    CNY = ByAlphabeticCode.map['CNY']
    COP = ByAlphabeticCode.map['COP']
    CRC = ByAlphabeticCode.map['CRC']
    CUC = ByAlphabeticCode.map['CUC']
    CUP = ByAlphabeticCode.map['CUP']
    CVE = ByAlphabeticCode.map['CVE']
    CZK = ByAlphabeticCode.map['CZK']
    DJF = ByAlphabeticCode.map['DJF']
    DKK = ByAlphabeticCode.map['DKK']
    DOP = ByAlphabeticCode.map['DOP']
    DZD = ByAlphabeticCode.map['DZD']
    EGP = ByAlphabeticCode.map['EGP']
    ERN = ByAlphabeticCode.map['ERN']
    ETB = ByAlphabeticCode.map['ETB']
    EUR = ByAlphabeticCode.map['EUR']
    FJD = ByAlphabeticCode.map['FJD']
    FKP = ByAlphabeticCode.map['FKP']
    GBP = ByAlphabeticCode.map['GBP']
    GEL = ByAlphabeticCode.map['GEL']
    GHS = ByAlphabeticCode.map['GHS']
    GIP = ByAlphabeticCode.map['GIP']
    GMD = ByAlphabeticCode.map['GMD']
    GNF = ByAlphabeticCode.map['GNF']
    GTQ = ByAlphabeticCode.map['GTQ']
    GYD = ByAlphabeticCode.map['GYD']
    HKD = ByAlphabeticCode.map['HKD']
    HNL = ByAlphabeticCode.map['HNL']
    HTG = ByAlphabeticCode.map['HTG']
    HUF = ByAlphabeticCode.map['HUF']
    IDR = ByAlphabeticCode.map['IDR']
    ILS = ByAlphabeticCode.map['ILS']
    INR = ByAlphabeticCode.map['INR']
    IQD = ByAlphabeticCode.map['IQD']
    IRR = ByAlphabeticCode.map['IRR']
    ISK = ByAlphabeticCode.map['ISK']
    JMD = ByAlphabeticCode.map['JMD']
    JOD = ByAlphabeticCode.map['JOD']
    JPY = ByAlphabeticCode.map['JPY']
    KES = ByAlphabeticCode.map['KES']
    KGS = ByAlphabeticCode.map['KGS']
    KHR = ByAlphabeticCode.map['KHR']
    KMF = ByAlphabeticCode.map['KMF']
    KPW = ByAlphabeticCode.map['KPW']
    KRW = ByAlphabeticCode.map['KRW']
    KWD = ByAlphabeticCode.map['KWD']
    KYD = ByAlphabeticCode.map['KYD']
    KZT = ByAlphabeticCode.map['KZT']
    LAK = ByAlphabeticCode.map['LAK']
    LBP = ByAlphabeticCode.map['LBP']
    LKR = ByAlphabeticCode.map['LKR']
    LRD = ByAlphabeticCode.map['LRD']
    LSL = ByAlphabeticCode.map['LSL']
    LYD = ByAlphabeticCode.map['LYD']
    MAD = ByAlphabeticCode.map['MAD']
    MDL = ByAlphabeticCode.map['MDL']
    MGA = ByAlphabeticCode.map['MGA']
    MKD = ByAlphabeticCode.map['MKD']
    MMK = ByAlphabeticCode.map['MMK']
    MNT = ByAlphabeticCode.map['MNT']
    MOP = ByAlphabeticCode.map['MOP']
    MRU = ByAlphabeticCode.map['MRU']
    MUR = ByAlphabeticCode.map['MUR']
    MVR = ByAlphabeticCode.map['MVR']
    MWK = ByAlphabeticCode.map['MWK']
    MXN = ByAlphabeticCode.map['MXN']
    MYR = ByAlphabeticCode.map['MYR']
    MZN = ByAlphabeticCode.map['MZN']
    NAD = ByAlphabeticCode.map['NAD']
    NGN = ByAlphabeticCode.map['NGN']
    NIO = ByAlphabeticCode.map['NIO']
    NOK = ByAlphabeticCode.map['NOK']
    NPR = ByAlphabeticCode.map['NPR']
    NZD = ByAlphabeticCode.map['NZD']
    OMR = ByAlphabeticCode.map['OMR']
    PAB = ByAlphabeticCode.map['PAB']
    PEN = ByAlphabeticCode.map['PEN']
    PGK = ByAlphabeticCode.map['PGK']
    PHP = ByAlphabeticCode.map['PHP']
    PKR = ByAlphabeticCode.map['PKR']
    PLN = ByAlphabeticCode.map['PLN']
    PYG = ByAlphabeticCode.map['PYG']
    QAR = ByAlphabeticCode.map['QAR']
    RON = ByAlphabeticCode.map['RON']
    RSD = ByAlphabeticCode.map['RSD']
    RUB = ByAlphabeticCode.map['RUB']
    RWF = ByAlphabeticCode.map['RWF']
    SAR = ByAlphabeticCode.map['SAR']
    SBD = ByAlphabeticCode.map['SBD']
    SCR = ByAlphabeticCode.map['SCR']
    SDG = ByAlphabeticCode.map['SDG']
    SEK = ByAlphabeticCode.map['SEK']
    SGD = ByAlphabeticCode.map['SGD']
    SHP = ByAlphabeticCode.map['SHP']
    SLE = ByAlphabeticCode.map['SLE']
    SLL = ByAlphabeticCode.map['SLL']
    SOS = ByAlphabeticCode.map['SOS']
    SRD = ByAlphabeticCode.map['SRD']
    SSP = ByAlphabeticCode.map['SSP']
    STN = ByAlphabeticCode.map['STN']
    SVC = ByAlphabeticCode.map['SVC']
    SYP = ByAlphabeticCode.map['SYP']
    SZL = ByAlphabeticCode.map['SZL']
    THB = ByAlphabeticCode.map['THB']
    TJS = ByAlphabeticCode.map['TJS']
    TMT = ByAlphabeticCode.map['TMT']
    TND = ByAlphabeticCode.map['TND']
    TOP = ByAlphabeticCode.map['TOP']
    TRY = ByAlphabeticCode.map['TRY']
    TTD = ByAlphabeticCode.map['TTD']
    TWD = ByAlphabeticCode.map['TWD']
    TZS = ByAlphabeticCode.map['TZS']
    UAH = ByAlphabeticCode.map['UAH']
    UGX = ByAlphabeticCode.map['UGX']
    USD = ByAlphabeticCode.map['USD']
    UYU = ByAlphabeticCode.map['UYU']
    UYW = ByAlphabeticCode.map['UYW']
    UZS = ByAlphabeticCode.map['UZS']
    VED = ByAlphabeticCode.map['VED']
    VES = ByAlphabeticCode.map['VES']
    VND = ByAlphabeticCode.map['VND']
    VUV = ByAlphabeticCode.map['VUV']
    WST = ByAlphabeticCode.map['WST']
    XAF = ByAlphabeticCode.map['XAF']
    XCD = ByAlphabeticCode.map['XCD']
    XOF = ByAlphabeticCode.map['XOF']
    XPF = ByAlphabeticCode.map['XPF']
    YER = ByAlphabeticCode.map['YER']
    ZAR = ByAlphabeticCode.map['ZAR']
    ZMW = ByAlphabeticCode.map['ZMW']
    ZWL = ByAlphabeticCode.map['ZWL']


# 数据完整性和一致性检验
ByAlphabeticCode.check(obj=Currency)



# -*- coding: utf-8 -*-
import os
import zipfile
import zstandard
from typing import IO, AnyStr, Iterator, Tuple, List


class TextReader:
    """多行文本分块读取器"""

    @classmethod
    def iter_zst_text(cls, f: AnyStr | IO, size: int = 64 * 1024 ** 2) -> Iterator[Tuple[str, str | None]]:
        path: str | None = None
        if isinstance(f, str):
            assert os.path.isfile(f)
            path: str = f
            fd: IO = open(f, mode='rb')
        else:
            fd: IO = f
        try:
            decompressor = zstandard.ZstdDecompressor().stream_reader(f)
            for text in cls.iter_stream_text(stream=decompressor, size=size):
                yield text, path
        finally:
            fd.close()

    @classmethod
    def iter_zip_text(cls, f: AnyStr | IO, size: int = 64 * 1024 ** 2) -> Iterator[Tuple[str, str | None]]:
        if isinstance(f, str):
            assert os.path.isfile(f)
        with zipfile.ZipFile(f) as zf:
            for zi in zf.infolist():
                if zi.is_dir():
                    continue
                with zf.open(zi) as stream:
                    for text in cls.iter_stream_text(stream=stream, size=size):
                        yield text, zi.filename

    @classmethod
    def iter_stream_text(cls, stream: IO, size: int = 64 * 1024 ** 2, encodings: Tuple = ('utf-8', 'gbk')) -> Iterator[str]:
        for payload in cls.iter_stream_bytes(stream=stream, size=size):
            for i, encoding in enumerate(encodings):
                try:
                    text: str = payload.decode(encoding)
                    yield text
                    break
                except Exception as e:
                    if i == len(encodings) - 1:
                        raise e

    @classmethod
    def iter_stream_bytes(cls, stream: IO, size: int = 64 * 1024 ** 2) -> Iterator[bytes]:
        tail: bytes = b''
        i: int = 0
        while segment := stream.read(size):  # 每次处理size大小的多行文本
            idx: int = segment.rfind(b'\n')
            payload: bytes = tail + segment[:idx + 1]
            tail: bytes = segment[idx + 1:]
            yield payload
            i += 1



"""
Python SDK for jyfund
"""
# 枚举类
from .bank import Bank
from .country import CountryA2, CountryA3
from .currency import Currency
from .exchange import Exchange
from .lock import AsyncRWLock, SyncRWLock
from .symbol import (
    BlockType,
    BondType,
    DrType,
    FundType,
    FuturesType,
    IndexType,
    OptionsType,
    SpotType,
    StockType,
    SymbolFlag,
    SymbolSubType,
    SymbolType,
    WarrantsType,
)
from .symbol_em import EMParser, KLineRecord, TrendRecord

__version__ = "0.8.0"



# -*- coding: utf-8 -*-
"""国内外交易品种类型枚举"""
import functools
import re
from dataclasses import dataclass
from enum import Enum
from typing import Dict, FrozenSet, List, Tuple
from jy_lib import Exchange


class SymbolSubType(Enum):
    def parent(self) -> 'SymbolType':
        """获取当前的父类型"""
        return SymbolType(self.__class__)


class StockType(SymbolSubType):
    """股票"""
    STOCK_A = '以人民币交易的股票（主板）'  # 主板A股 SH:ASH
    STOCK_B = '以美元交易的股票'  # 主板B股 SH:BSH
    STOCK_K = '以人民币交易的股票（科创板）'  # 科创板 SH:KSH
    STOCK_G = '创业板'  # 创业板GEM SZ:GEM
    STOCK_P = '优先股'  # SH:OPS,PPS SZ:33
    STOCK_OTHER = '其它股票'  # SH:OEQ,OPS,PPS


class FundType(SymbolSubType):
    """基金"""
    FUND_CLOSE = '封闭式基金'  # SH:CEF
    FUND_OPEN = '开放式基金'  # SH:OEF
    FUND_ETF = '交易所交易基金'  # SH:EBS
    FUND_LOF = 'LOF基金'  # SH:LOF
    FUND_REITS = '公募REITs'  # SH:RET
    FUND_STOCK = '股票基金'
    FUND_BOND = '债券型基金'
    FUND_DERIVATIVES = '衍生品基金'
    FUND_HYBRID = '混合型基金'
    FUND_FOF = 'FOF基金'
    FUND_OTHER = '其它基金'  # SH:OFN


class BondType(SymbolSubType):
    """债券"""
    BOND_NATIONAL = '国债、地方债、政府支持债、政策性金融债'  # SH:GBF
    BOND_ENTERPRISE = '企业债'  # SH:CBF
    BOND_CONVERTIBLE = '可转换企业债'  # SH:CCF
    BOND_EXCHANGEABLE = '可交换债'
    BOND_COMPANY = '公司债、企业债、可交换债、政府支持债'  # SH:CPF
    BOND_BUYBACK = '质押式回购'  # SH:CRP SZ:12
    BOND_ABS = '资产支持证券'
    BOND_OTHER = '其它债券'  # SH:DST,DVP,OBD,WIT,QRP,AMP,TCB


class OptionsType(SymbolSubType):
    """期权"""
    OPTIONS_STOCK = '个股期权'  # SZ:29
    OPTIONS_ETF = 'ETF期权'  # SZ:30
    OPTIONS_INDEX = '指数期权'  # CFFEX
    OPTIONS_COMMODITY = '商品期权'
    OPTIONS_FOREX = '外汇期权'
    OPTIONS_OTHER = '其他期权'


class DrType(SymbolSubType):
    """存托凭证"""
    DR_GDR = '全球存托凭证'
    DR_OTHER = '其他存托凭证'


class IndexType(SymbolSubType):
    """指数"""
    INDEX_OTHER = '其他指数'


class FuturesType(SymbolSubType):
    """期货"""
    FUTURES_STOCK = '股票期货'
    FUTURES_INDEX = '指数期货'
    FUTURES_COMMODITY = '商品期货'
    FUTURES_BOND = '债券期货'
    FUTURES_OTHER = '其他期货'


class WarrantsType(SymbolSubType):
    """权证"""
    WARRANTS_CBBC = '牛熊证'
    WARRANTS_INLINE = '界内证'
    WARRANTS_OTHER = '其他权证'


class BlockType(SymbolSubType):
    """板块"""
    BLOCK_CONCEPT = '概念板块'
    BLOCK_REGION = '地域板块'
    BLOCK_INDUSTRY = '行业板块'
    BLOCK_OTHER = '其他板块'


class SpotType(SymbolSubType):
    """现货"""
    SPOT_COMMODITY = '商品现货'
    SPOT_OTHER = '其他现货'


class SymbolType(Enum):
    """基础类别"""
    STOCK = StockType
    BOND = BondType
    FUND = FundType
    INDEX = IndexType
    FUTURES = FuturesType
    OPTIONS = OptionsType
    WARRANTS = WarrantsType
    DR = DrType
    SPOT = SpotType
    BLOCK = BlockType


@dataclass
class SymbolFlag:
    is_st: bool | None = None                       # 风险警示=True
    is_new: bool | None = None                      # 新股=True
    is_approval: bool | None = None                 # 核准制=True
    is_registration: bool | None = None             # 注册制=True
    is_base_market: bool | None = None              # 基础层=True
    is_innovation_market: bool | None = None        # 创新层=True
    is_call_auction: bool | None = None             # 集合竞价=True
    is_market_making: bool | None = None            # 做市=True
    is_pt: bool | None = None                       # 两网及退市=True
    is_call: bool | None = None                     # 看涨=True
    is_put: bool | None = None                      # 看跌=True
    is_northbound: bool | None = None               # 北向=True
    is_southbound: bool | None = None               # 南向=True
    is_main_contract: bool | None = None            # 主力合约=True
    is_second_main_contract: bool | None = None     # 次主力合约=True
    is_main_perpetual_contract: bool | None = None          # 主力永续合约=True
    is_second_main_perpetual_contract: bool | None = None   # 次主力永续合约=True


class SymbolDetail:
    def __init__(self, original_symbol_id: str, exchange: Exchange, instrument_id: str, instrument_type: SymbolType,
                 instrument_name: str = '', product_id: str = '', underlying_symbol: 'SymbolDetail' = None,
                 expired: bool = False
                 ):
        self.original_symbol_id: str = original_symbol_id
        self.exchange: Exchange = exchange
        self.instrument_id: str = instrument_id
        self.symbol_id: str = f'{instrument_id}.{exchange.value.acronym}'
        self.instrument_type: SymbolType = instrument_type
        self.instrument_name: str = instrument_name
        self.product_id: str = product_id      # 期货/期权合约对应的产品
        self.underlying_symbol: SymbolDetail | None = underlying_symbol   # 期权合约对应的期货合约
        self.expired: bool = expired

    def __repr__(self):
        return f'{self.__class__.__name__}({self.symbol_id}, {self.instrument_type.name}, {self.instrument_name}, {self.product_id})'


# 国内期货交易所
DomesticFuturesExchanges: FrozenSet[Exchange] = frozenset({
    Exchange.CZCE, Exchange.DCE, Exchange.SHFE, Exchange.INE, Exchange.CFFEX, Exchange.GFEX,
})


# 国内期权交易所
DomesticOptionsExchanges: FrozenSet[Exchange] = frozenset({
    Exchange.CZCE, Exchange.DCE, Exchange.SHFE, Exchange.INE, Exchange.CFFEX, Exchange.GFEX,
    Exchange.SSE, Exchange.SZSE,
})


class SymbolGuesser:
    """代码猜解器"""

    @classmethod
    @functools.lru_cache(maxsize=2**16)
    def guess_symbol_detail(cls, original_symbol_id: str) -> SymbolDetail | None:
        index: int = original_symbol_id.find('.')
        if index <= 0:
            return
        a: str = original_symbol_id[:index]
        b: str = original_symbol_id[index + 1:]
        eb: Exchange | None = Exchange.find(market_id=b)
        ea: Exchange | None = Exchange.find(market_id=a)
        if eb and ea:
            return
        elif eb:
            iid: str = a
            exchange: Exchange = eb
        elif ea:
            iid: str = b
            exchange: Exchange = ea
        else:
            return
        l_iid: int = len(iid)
        if exchange == Exchange.SSE:
            if l_iid == 8 and iid.startswith('1'):
                instrument_type = SymbolType.OPTIONS
            elif l_iid == 6 and iid.startswith('000'):
                instrument_type = SymbolType.INDEX
            elif l_iid == 6 and iid.startswith('5'):
                instrument_type = SymbolType.FUND
            elif l_iid == 6 and iid.startswith(('6', '9')):
                instrument_type = SymbolType.STOCK
            elif l_iid == 6 and iid.startswith(('100', '11', '204', '1260')):
                instrument_type = SymbolType.BOND
            else:
                return
            return SymbolDetail(
                original_symbol_id=original_symbol_id,
                exchange=exchange,
                instrument_id=iid,
                instrument_type=instrument_type,
            )
        elif exchange == Exchange.SZSE:
            if len(iid) == 8 and iid.startswith('9'):
                instrument_type = SymbolType.OPTIONS
            elif l_iid == 6 and iid.startswith('399'):
                instrument_type = SymbolType.INDEX
            elif l_iid == 6 and iid.startswith(('15', '16', '18')):
                instrument_type = SymbolType.FUND
            elif l_iid == 6 and iid.startswith(('2', '0', '3')):
                instrument_type = SymbolType.STOCK
            elif l_iid == 6 and iid.startswith(('12', '131', '1150')):
                instrument_type = SymbolType.BOND
            else:
                return
            return SymbolDetail(
                original_symbol_id=original_symbol_id,
                exchange=exchange,
                instrument_id=iid,
                instrument_type=instrument_type,
            )
        elif exchange == Exchange.BSE:
            if l_iid == 6 and iid.startswith('899'):
                instrument_type = SymbolType.INDEX
            elif l_iid == 6 and iid.startswith('82'):
                instrument_type = SymbolType.BOND
            elif l_iid == 6 and iid.startswith(('43', '83', '87')):
                instrument_type = SymbolType.STOCK
            else:
                return
            return SymbolDetail(
                original_symbol_id=original_symbol_id,
                exchange=exchange,
                instrument_id=iid,
                instrument_type=instrument_type,
            )
        elif exchange == Exchange.HKEX:
            if l_iid == 5:
                match iid[0]:
                    case '0' | '8':         # 港股主板
                        instrument_type = SymbolType.STOCK
                    case '1' | '2':         # 港股涡轮
                        instrument_type = SymbolType.WARRANTS
                    case '4' | '5' | '6':   # 港股牛熊证
                        instrument_type = SymbolType.OPTIONS
                    case _:
                        return
                return SymbolDetail(
                    original_symbol_id=original_symbol_id,
                    exchange=exchange,
                    instrument_id=iid,
                    instrument_type=instrument_type,
                )
        elif exchange in DomesticFuturesExchanges:
            # 不管期货还是期权 先分交易所大小写规格化一下
            match exchange:
                case Exchange.SHFE | Exchange.INE | Exchange.DCE | Exchange.GFEX:
                    code = iid.lower()
                case Exchange.CZCE | Exchange.CFFEX:
                    code = iid.upper()
                case _:
                    code = iid

            # 尝试匹配 期货
            match exchange:
                case Exchange.CZCE:
                    rs: List[Tuple[str, str]] = re.findall(r'^([A-Z]+)(\d{3,4})$', code, re.IGNORECASE)
                case _:
                    rs: List[Tuple[str, str]] = re.findall(r'^([A-Z]+)(\d{4})$', code, re.IGNORECASE)
            if rs:
                product_id, ym = rs[0]
                if exchange == Exchange.CZCE:
                    ym: str = ym[-3:]
                if 1 <= int(ym) % 100 <= 12:
                    return SymbolDetail(
                        original_symbol_id=original_symbol_id,
                        exchange=exchange,
                        instrument_id=f'{product_id}{ym}',
                        instrument_type=SymbolType.FUTURES,
                        product_id=product_id,
                    )

            # 尝试匹配 期货加权
            rs: List[Tuple[str, str]] = re.findall(r'^([a-zA-Z]+)(JQ00|8888)$', code, re.IGNORECASE)
            if rs:
                product_id, _ = rs[0]
                return SymbolDetail(
                    original_symbol_id=original_symbol_id,
                    exchange=exchange,
                    instrument_id=f'{product_id}8888',
                    instrument_type=SymbolType.FUTURES,
                    product_id=product_id,
                )

            # 尝试匹配 期货主连
            rs = re.findall(r'^([a-zA-Z]+)(00|9999)$', code, re.IGNORECASE)
            if rs:
                product_id, _ = rs[0]
                return SymbolDetail(
                    original_symbol_id=original_symbol_id,
                    exchange=exchange,
                    instrument_id=f'{product_id}9999',
                    instrument_type=SymbolType.FUTURES,
                    product_id=product_id,
                )

            # 尝试匹配 期权
            match exchange:
                case Exchange.CZCE:
                    rs: List[Tuple[str, str, str, str]] = re.findall(
                        r'^([A-Z]+?)(\d{3,4})-?([CP])-?(\d+)$', code, re.IGNORECASE)
                case _:
                    rs: List[Tuple[str, str, str, str]] = re.findall(
                        r'^([A-Z]+?)(\d{4})-?([CP])-?(\d+)$', code, re.IGNORECASE)
            if rs:
                product_id, ym, cp, ep = rs[0]
                if exchange == Exchange.CZCE:
                    ym: str = ym[-3:]
                if 1 <= int(ym) % 100 <= 12:
                    cp: str = cp.upper()   # C / P
                    match exchange:
                        case Exchange.DCE | Exchange.GFEX:
                            # y2407-C-8700 si2408-C-16200
                            code = f'{product_id.lower()}{ym}-{cp}-{ep}'
                        case Exchange.SHFE | Exchange.INE:
                            # sc2406C640
                            code = f'{product_id.lower()}{ym}{cp}{ep}'
                        case Exchange.CZCE:
                            # AP410C7300
                            code = f'{product_id.upper()}{ym}{cp}{ep}'
                        case Exchange.CFFEX:
                            # MO2406-C-3800
                            code = f'{product_id.upper()}{ym}-{cp}-{ep}'
                        case _:
                            code = code
                    return SymbolDetail(
                        original_symbol_id=original_symbol_id,
                        exchange=exchange,
                        instrument_id=code,
                        instrument_type=SymbolType.OPTIONS,
                        product_id=product_id,
                        underlying_symbol=cls.guess_symbol_detail(
                            original_symbol_id=f'{product_id}{ym}.{exchange.value.acronym}'
                        ),
                    )

    @classmethod
    def test(cls):
        """验证"""
        # 上交所
        sd = cls.guess_symbol_detail(original_symbol_id='600000.SH')
        assert sd.symbol_id == '600000.SSE'
        assert sd.instrument_type == SymbolType.STOCK

        sd = cls.guess_symbol_detail(original_symbol_id='688192.SH')
        assert sd.symbol_id == '688192.SSE'
        assert sd.instrument_type == SymbolType.STOCK

        sd = cls.guess_symbol_detail(original_symbol_id='900906.SH')
        assert sd.symbol_id == '900906.SSE'
        assert sd.instrument_type == SymbolType.STOCK

        sd = cls.guess_symbol_detail(original_symbol_id='10006167.SHO')
        assert sd.symbol_id == '10006167.SSE'
        assert sd.instrument_type == SymbolType.OPTIONS

        sd = cls.guess_symbol_detail(original_symbol_id='10006168.SH')
        assert sd.symbol_id == '10006168.SSE'
        assert sd.instrument_type == SymbolType.OPTIONS

        sd = cls.guess_symbol_detail(original_symbol_id='510300.SH')
        assert sd.symbol_id == '510300.SSE'
        assert sd.instrument_type == SymbolType.FUND

        sd = cls.guess_symbol_detail(original_symbol_id='508018.SH')
        assert sd.symbol_id == '508018.SSE'
        assert sd.instrument_type == SymbolType.FUND

        sd = cls.guess_symbol_detail(original_symbol_id='204001.SH')
        assert sd.symbol_id == '204001.SSE'
        assert sd.instrument_type == SymbolType.BOND

        sd = cls.guess_symbol_detail(original_symbol_id='110059.SH')
        assert sd.symbol_id == '110059.SSE'
        assert sd.instrument_type == SymbolType.BOND

        sd = cls.guess_symbol_detail(original_symbol_id='000001.SH')
        assert sd.symbol_id == '000001.SSE'
        assert sd.instrument_type == SymbolType.INDEX

        # 深交所
        sd = cls.guess_symbol_detail(original_symbol_id='000001.SZ')
        assert sd.symbol_id == '000001.SZSE'
        assert sd.instrument_type == SymbolType.STOCK

        sd = cls.guess_symbol_detail(original_symbol_id='300282.SZ')
        assert sd.symbol_id == '300282.SZSE'
        assert sd.instrument_type == SymbolType.STOCK

        sd = cls.guess_symbol_detail(original_symbol_id='002746.SZ')
        assert sd.symbol_id == '002746.SZSE'
        assert sd.instrument_type == SymbolType.STOCK

        sd = cls.guess_symbol_detail(original_symbol_id='200029.SZ')
        assert sd.symbol_id == '200029.SZSE'
        assert sd.instrument_type == SymbolType.STOCK

        sd = cls.guess_symbol_detail(original_symbol_id='90002795.SZO')
        assert sd.symbol_id == '90002795.SZSE'
        assert sd.instrument_type == SymbolType.OPTIONS

        sd = cls.guess_symbol_detail(original_symbol_id='90002796.SZ')
        assert sd.symbol_id == '90002796.SZSE'
        assert sd.instrument_type == SymbolType.OPTIONS

        sd = cls.guess_symbol_detail(original_symbol_id='159919.SZ')
        assert sd.symbol_id == '159919.SZSE'
        assert sd.instrument_type == SymbolType.FUND

        sd = cls.guess_symbol_detail(original_symbol_id='180301.SZ')
        assert sd.symbol_id == '180301.SZSE'
        assert sd.instrument_type == SymbolType.FUND

        sd = cls.guess_symbol_detail(original_symbol_id='131810.SZ')
        assert sd.symbol_id == '131810.SZSE'
        assert sd.instrument_type == SymbolType.BOND

        sd = cls.guess_symbol_detail(original_symbol_id='123008.SZ')
        assert sd.symbol_id == '123008.SZSE'
        assert sd.instrument_type == SymbolType.BOND

        sd = cls.guess_symbol_detail(original_symbol_id='399001.SZ')
        assert sd.symbol_id == '399001.SZSE'
        assert sd.instrument_type == SymbolType.INDEX

        # 北交所
        sd = cls.guess_symbol_detail(original_symbol_id='873122.BJ')
        assert sd.symbol_id == '873122.BSE'
        assert sd.instrument_type == SymbolType.STOCK

        sd = cls.guess_symbol_detail(original_symbol_id='430510.BJE')
        assert sd.symbol_id == '430510.BSE'
        assert sd.instrument_type == SymbolType.STOCK

        # 上期所
        sd = cls.guess_symbol_detail(original_symbol_id='fu2405.SF')
        assert sd.symbol_id == 'fu2405.SHFE'
        assert sd.instrument_type == SymbolType.FUTURES

        sd = cls.guess_symbol_detail(original_symbol_id='SC2406-C-520.SF')
        assert sd.symbol_id == 'sc2406C520.SHFE'
        assert sd.instrument_type == SymbolType.OPTIONS

        # 能源所
        sd = cls.guess_symbol_detail(original_symbol_id='BC2407.INE')
        assert sd.symbol_id == 'bc2407.INE'
        assert sd.instrument_type == SymbolType.FUTURES

        sd = cls.guess_symbol_detail(original_symbol_id='SC2406-C-570.INE')
        assert sd.symbol_id == 'sc2406C570.INE'
        assert sd.instrument_type == SymbolType.OPTIONS

        # 大商所
        sd = cls.guess_symbol_detail(original_symbol_id='A2405.DF')
        assert sd.symbol_id == 'a2405.DCE'
        assert sd.instrument_type == SymbolType.FUTURES

        sd = cls.guess_symbol_detail(original_symbol_id='A2407C4050.DCE')
        assert sd.symbol_id == 'a2407-C-4050.DCE'
        assert sd.instrument_type == SymbolType.OPTIONS

        # 郑商所
        sd = cls.guess_symbol_detail(original_symbol_id='Oi409.ZF')
        assert sd.symbol_id == 'OI409.CZCE'
        assert sd.instrument_type == SymbolType.FUTURES

        sd = cls.guess_symbol_detail(original_symbol_id='ap411-c-8900.CZCE')
        assert sd.symbol_id == 'AP411C8900.CZCE'
        assert sd.instrument_type == SymbolType.OPTIONS

        # 广期所
        sd = cls.guess_symbol_detail(original_symbol_id='si2410.GF')
        assert sd.symbol_id == 'si2410.GFEX'
        assert sd.instrument_type == SymbolType.FUTURES

        sd = cls.guess_symbol_detail(original_symbol_id='si2406P10600.GFEX')
        assert sd.symbol_id == 'si2406-P-10600.GFEX'
        assert sd.instrument_type == SymbolType.OPTIONS

        # 中金所
        sd = cls.guess_symbol_detail(original_symbol_id='t2406.IF')
        assert sd.symbol_id == 'T2406.CFFEX'
        assert sd.instrument_type == SymbolType.FUTURES

        sd = cls.guess_symbol_detail(original_symbol_id='HO2406C3000.CCFX')
        assert sd.symbol_id == 'HO2406-C-3000.CFFEX'
        assert sd.instrument_type == SymbolType.OPTIONS

        # 主连
        sd = cls.guess_symbol_detail(original_symbol_id='t00.IF')
        assert sd.symbol_id == 'T9999.CFFEX'
        assert sd.instrument_type == SymbolType.FUTURES

        sd = cls.guess_symbol_detail(original_symbol_id='t9999.IF')
        assert sd.symbol_id == 'T9999.CFFEX'
        assert sd.instrument_type == SymbolType.FUTURES

        # 指数
        sd = cls.guess_symbol_detail(original_symbol_id='tJQ00.IF')
        assert sd.symbol_id == 'T8888.CFFEX'
        assert sd.instrument_type == SymbolType.FUTURES

        sd = cls.guess_symbol_detail(original_symbol_id='t8888.IF')
        assert sd.symbol_id == 'T8888.CFFEX'
        assert sd.instrument_type == SymbolType.FUTURES

        # 港股
        sd = cls.guess_symbol_detail(original_symbol_id='HK.00700')
        assert sd.symbol_id == '00700.HKEX'
        assert sd.instrument_type == SymbolType.STOCK

        sd = cls.guess_symbol_detail(original_symbol_id='HK.29951')
        assert sd.symbol_id == '29951.HKEX'
        assert sd.instrument_type == SymbolType.WARRANTS

        sd = cls.guess_symbol_detail(original_symbol_id='HK.69994')
        assert sd.symbol_id == '69994.HKEX'
        assert sd.instrument_type == SymbolType.OPTIONS

        sd = cls.guess_symbol_detail(original_symbol_id='00700.HGT')
        assert sd.symbol_id == '00700.HKEX'
        assert sd.instrument_type == SymbolType.STOCK

        sd = cls.guess_symbol_detail(original_symbol_id='00700.SGT')
        assert sd.symbol_id == '00700.HKEX'
        assert sd.instrument_type == SymbolType.STOCK

        #
        assert not cls.guess_symbol_detail(original_symbol_id='t01.IF')
        assert not cls.guess_symbol_detail(original_symbol_id='t012.IF')
        assert not cls.guess_symbol_detail(original_symbol_id='tjq01.IF')
        assert not cls.guess_symbol_detail(original_symbol_id='000700.SGT')


if __name__ == '__main__':
    SymbolGuesser.test()



# -*- coding: utf-8 -*-
import asyncio
import threading
import time


class SyncRWLock:
    """同步读写锁,单写多读"""
    def __init__(self, reentrant: bool = True):
        self.readers: int = 0  # 读取者数量
        self.reentrant: bool = reentrant
        if reentrant:
            self.r_lock: threading.RLock = threading.RLock()
            self.w_lock: threading.RLock = threading.RLock()
        else:
            self.r_lock: threading.Lock = threading.Lock()
            self.w_lock: threading.Lock = threading.Lock()

    def read_acquire(self):
        """获取读锁"""
        with self.r_lock:
            self.readers += 1
            if self.readers == 1:
                self.w_lock.acquire()

    def read_release(self):
        """释放读锁"""
        with self.r_lock:
            self.readers -= 1
            if self.readers == 0:
                self.w_lock.release()

    def write_acquire(self):
        """获取写锁"""
        self.w_lock.acquire()

    def write_release(self):
        """释放写锁"""
        self.w_lock.release()

    def __enter__(self):
        """仅针对读锁的上下文"""
        self.read_acquire()
        return None

    def __exit__(self, exc_type, exc, tb):
        """仅针对读锁的上下文"""
        self.read_release()


class AsyncRWLock:
    """异步读写锁,单写多读"""
    def __init__(self):
        self.readers: int = 0  # 读取者数量
        self.r_lock: asyncio.Lock = asyncio.Lock()
        self.w_lock: asyncio.Lock = asyncio.Lock()

    async def read_acquire(self):
        """获取读锁"""
        async with self.r_lock:
            self.readers += 1
            if self.readers == 1:
                await self.w_lock.acquire()

    async def read_release(self):
        """释放读锁"""
        async with self.r_lock:
            self.readers -= 1
            if self.readers == 0:
                self.w_lock.release()

    async def write_acquire(self):
        """获取写锁"""
        await self.w_lock.acquire()

    async def write_release(self):
        """释放写锁"""
        self.w_lock.release()

    async def __aenter__(self):
        """仅针对读锁的上下文"""
        await self.read_acquire()
        return None

    async def __aexit__(self, exc_type, exc, tb):
        """仅针对读锁的上下文"""
        await self.read_release()



# -*- coding: utf-8 -*-
import signal
import threading
from loguru import logger
from types import FrameType


class KeyboardInterruptProtect:
    """
    防止代码段被Ctrl+C终止
    with KeyboardInterruptProtect():
        do()
    """

    @classmethod
    def is_main_thread(cls) -> bool:
        """当前线程是否是主线程"""
        return threading.current_thread().name == 'MainThread'

    @classmethod
    def signal_handler(cls, signum: int, frame: FrameType | None):
        logger.warning(f'Ignore KeyboardInterrupt')

    def __enter__(self):
        if self.is_main_thread():
            signal.signal(signal.SIGINT, self.signal_handler)

    def __exit__(self, exc_type, exc_val, exc_tb):
        # 重置signal处理函数为默认行为
        if self.is_main_thread():
            signal.signal(signal.SIGINT, signal.SIG_DFL)



# -*- coding: utf-8 -*-
import time


class Time:
    # 多次取样计算基准时间
    _reference_time = min([int(time.time() * 10 ** 9) - time.perf_counter_ns() for _ in range(10000)])

    @classmethod
    def time_ns(cls) -> int:
        """获取纳秒时间戳"""
        return cls._reference_time + time.perf_counter_ns()

    @classmethod
    def time_us(cls) -> int:
        """获取微秒时间戳"""
        return cls.time_ns() // 1000

    @classmethod
    def time_ms(cls) -> int:
        """获取毫秒时间戳"""
        return cls.time_ns() // 1000_000



# -*- coding: utf-8 -*-
import collections
import datetime
import threading
from abc import abstractmethod
import clickhouse_driver
import io
import os
import polars as pl
import requests
import time
import urllib.parse
import zstandard
from typing import Dict, Type, List, Any, Tuple, DefaultDict
from loguru import logger
from jy_lib.compressor import ZstdStreamCompressor
from jy_lib.interrupt_protect import KeyboardInterruptProtect


class FieldDetail:
    def __init__(self, name: str, read_dtype: Any = pl.Float64, dtype: Any = None,
                 convertor: pl.Expr = None, comment: str = '', drop: bool = False):
        self.name: str = name                           # 字段名
        self.read_dtype: Any = read_dtype              # 从csv读入pl的数据类型
        self.dtype: Type | pl.Datetime = dtype or self.read_dtype     # 从csv读入pl的数据类型
        if convertor is not None:
            self.convertor: pl.Expr = convertor
        elif read_dtype != self.dtype:
            self.convertor = pl.col(name).cast(self.dtype)
        else:
            self.convertor = None
        self.comment: str = comment                     # 备注
        self.drop: bool = drop                          # 是否需要在数据库中丢弃该列

    @property
    def ch_type(self) -> str:
        """clickhouse中对应列的数据类型"""
        if self.dtype == pl.Datetime(time_unit='ms'):
            return 'DateTime64(3)'
        elif self.dtype == pl.Datetime(time_unit='us'):
            return 'DateTime64(6)'
        elif self.dtype == pl.Datetime(time_unit='ns'):
            return 'DateTime64(9)'
        else:
            return self.dtype.base_type().__name__

    @property
    def ch_codec(self) -> str:
        """clickhouse中对应列的编码类型"""
        if self.dtype.base_type() == pl.Datetime:
            codec = 'CODEC(DoubleDelta, ZSTD(9))'
        else:
            codec = 'CODEC(ZSTD(9))'
        return codec

    @property
    def ch_comment(self) -> str:
        """clickhouse中对应列的注释"""
        return f"COMMENT '{self.comment}'"

    def ch_column(self, nullable: bool = True) -> str:
        """clickhouse中对应列的创建语句"""
        ch_type: str = f'Nullable({self.ch_type})' if nullable else self.ch_type
        return f"`{self.name}` {ch_type} {self.ch_comment} {self.ch_codec}"


class DefaultFieldDetails:
    def __init__(self):
        self.fds: Dict[str, FieldDetail] = {}

    def add(self, fd: FieldDetail | List[FieldDetail]):
        if isinstance(fd, FieldDetail):
            fd = [fd]
        assert isinstance(fd, list)
        for _fd in fd:
            assert isinstance(_fd, FieldDetail)
            assert _fd.name not in self.fds
            self.fds[_fd.name] = _fd

    def get(self, name: str) -> FieldDetail:
        return self.fds.get(name)

    def __getitem__(self, item):
        return self.fds[item]


class QuoteStruct:
    subclasses: List = []
    fields: List[FieldDetail] = []
    priority_cols: List[str] = []

    def __class_getitem__(cls, item) -> FieldDetail:
        for fd in cls.fields:
            if fd.name == item:
                return fd
        raise ValueError()

    @classmethod
    def get(cls, item: str) -> FieldDetail:
        for fd in cls.fields:
            if fd.name == item:
                return fd

    @classmethod
    def read_dtypes(cls) -> Dict[str, Type]:
        """获取polars从csv加载时的类型"""
        return {fd.name: fd.read_dtype for fd in cls.fields}

    @classmethod
    def read_columns(cls) -> List[str]:
        """获取polars从csv加载时的列名列表"""
        return [fd.name for fd in cls.fields]

    @classmethod
    def convertors(cls, symbol: str = None) -> List[pl.Expr]:
        """polars的列类型转换"""
        pk1: str = cls.pk1()
        convertors: List[pl.Expr] = [fd.convertor for fd in cls.fields if fd.convertor is not None]
        for fd in cls.fields:
            if fd.name == pk1:
                if symbol:
                    convertors.append(pl.lit(symbol).alias(pk1))
                break
        else:
            assert symbol
            convertors.append(pl.lit(symbol).alias(pk1))
        return convertors

    @classmethod
    @abstractmethod
    def table_name(cls, *args, **kwargs) -> str:
        pass

    @classmethod
    def convertor_pk1(cls, symbol: str) -> pl.Expr:
        """polars添加symbol列"""
        return pl.lit(symbol).alias(cls.pk1())

    @classmethod
    def cols(cls) -> List[str]:
        return cls.priority_cols + [fd.name for fd in cls.fields if fd.name and fd.name not in cls.priority_cols]

    @classmethod
    def pk1(cls) -> str:
        return cls.priority_cols[0]

    @classmethod
    def pk2(cls) -> str:
        return cls.priority_cols[1]

    @classmethod
    def sql_create_table(cls, fds: DefaultFieldDetails, period: str = None):
        """建表语句"""
        cols: List[str] = []
        for i, name in enumerate(cls.cols()):
            fd: FieldDetail | None = fds.get(name) or cls.get(name)
            if not fd:
                assert i == 0, '非主键列不能为空'
                fd = FieldDetail(name=cls.pk1(), read_dtype=pl.String, dtype=pl.String)
            if i == 0:
                assert fd.dtype == pl.String
            elif i == 1:
                assert fd.dtype.is_integer() or fd.dtype.base_type() == pl.Datetime
            if not fd.drop:
                cols.append(fd.ch_column(nullable=name not in cls.priority_cols))
        tb: str = '\n' + ',\n'.join(cols)
        sql: str = f"""
            CREATE TABLE IF NOT EXISTS `{cls.table_name(period=period)}` ({tb})
            ENGINE = ReplacingMergeTree()
            PARTITION BY sipHash64(`{cls.pk1()}`) % 256
            PRIMARY KEY (`{cls.pk1()}`, `{cls.pk2()}`)
            ORDER BY (`{cls.pk1()}`, `{cls.pk2()}`)"""
        return sql


class KlinesStruct(QuoteStruct):
    """K线数据"""
    fields: List[FieldDetail] = []
    priority_cols: List[str] = ['symbol', 'datetime']

    @classmethod
    def table_name(cls, period: str | None = None) -> str:
        assert period
        return f'{cls.__name__}_{period}'


class TickStruct(QuoteStruct):
    """Tick数据"""
    fields: List[FieldDetail] = []
    priority_cols: List[str] = ['symbol', 'datetime']

    @classmethod
    def table_name(cls, period: str | None = None) -> str:
        return cls.__name__


class TradeStruct(QuoteStruct):
    """Trade数据"""
    fields: List[FieldDetail] = []
    priority_cols: List[str] = ['symbol', 'id']

    @classmethod
    def table_name(cls, period: str | None = None) -> str:
        return cls.__name__


class ClickHouse:
    def __init__(self, url: str, http_port: int = 8123):
        self.client = clickhouse_driver.Client.from_url(url=url)
        self.http_port: int = http_port
        sp = urllib.parse.urlsplit(url)
        self.database_name: str = os.path.basename(sp.path)
        assert not sp.path or sp.path == f'/{self.database_name}'
        self.url_http: str = f'http://{sp.hostname}:{http_port}'
        self.credential: Dict = {
            'user': sp.username,
            'password': sp.password,
        }
        self.table_locks: DefaultDict[str, threading.Lock] = collections.defaultdict(threading.Lock)
        self.execute = self.client.execute

    def bulk_insert(self, table_name: str, df: pl.DataFrame):
        columns: str = ', '.join([f'`{column}`' for column in df.columns])
        sql: str = f'INSERT INTO `{table_name}`({columns}) VALUES'
        with KeyboardInterruptProtect():
            logger.debug(f'[{table_name}]准备插入{df.height:12,d}条数据')
            t1: float = time.time()
            self.client.execute(query=sql, params=df.to_dicts())
            t2: float = time.time()
            logger.success(f'[{table_name}]成功插入{df.height:12,d}条数据 耗时{t2 - t1:0.0f}秒')

    def import_csv(self, table_name: str, csv: str, height: int = None):
        sql: str = f'INSERT INTO `{self.database_name}`.`{table_name}` FORMAT CSVWithNames'
        params: Dict = {
            'query': sql,
            **self.credential,
            'enable_http_compression': 1,
        }
        t1: float = time.time()
        if height is None:
            height: int = -1
            sio = io.StringIO(csv)
            while sio.readline():
                height += 1
        if not csv:
            logger.warning(f'{sql}: empty csv')
            return
        assert height >= 0
        csv: bytes = csv.encode('utf-8')
        with ZstdStreamCompressor(size=len(csv), level=zstandard.STRATEGY_FAST) as sc:
            sc.update(data=csv)
            stream: io.BytesIO = sc.stream
        with KeyboardInterruptProtect():
            r = requests.post(
                url=self.url_http,
                params=params,
                data=stream.getvalue(),
                headers={
                    'Content-Encoding': 'zstd',
                    'Accept-Encoding': 'zstd',
                },
            )
            t2: float = time.time()
        assert r.status_code == 200, f'[{r.status_code}]{r.text}'
        if r.text:
            logger.info(r.text)
        logger.success(f'[{table_name}]成功插入{height:12,d}条数据 耗时{t2 - t1:0.0f}秒')

    def import_df(self, table_name: str, df: pl.DataFrame, pks: Tuple):
        """将polars.DataFrame导入ClickHouse,约束列为pks"""
        datetime_format: str = '%Y-%m-%d %H:%M:%S.%f'
        assert self.database_name
        assert pks and set(df.columns).issuperset(pks)
        # 按主键去重
        df = df.unique(subset=pks)
        logger.debug(f'[{table_name}]准备插入{df.height:12,d}条数据')
        # 通常第1主键为symbol
        pk1: str = pks[0]
        idx1: int = df.columns.index(pk1)
        assert df.dtypes[idx1] == pl.String
        pk1_vs = df[pk1].unique()
        outputs: str = ', '.join([f'`{pk}`' for pk in pks])
        sql = f"SELECT {outputs} FROM `{table_name}`"
        if pk1_vs.len() == 1:
            sql += f" WHERE `{pk1}` = '{pk1_vs.to_list()[0]}' "
        else:
            if pk1_vs.len() <= 1024:
                sql += f" WHERE `{pk1}` IN {pk1_vs.to_list()} "
        if len(pks) == 2:
            # 通常第2主键为id或datetime
            pk2: str = pks[1]
            idx2: int = df.columns.index(pk2)
            if df.dtypes[idx2].is_integer():
                pk2_min: int = df[pk2].min()
                pk2_max: int = df[pk2].max()
                sql += f"AND `{pk2}` BETWEEN {pk2_min} AND {pk2_max}"
            elif df.dtypes[idx2].base_type() == pl.Datetime:
                pk2_min: datetime.datetime = df[pk2].min()
                pk2_min: str = pk2_min.strftime(f'{datetime_format}')
                pk2_max: datetime.datetime = df[pk2].max()
                pk2_max: str = pk2_max.strftime(f'{datetime_format}')
                sql += f"AND `{pk2}` BETWEEN '{pk2_min}' AND '{pk2_max}'"
            elif df.dtypes[idx2] == pl.String:
                sql += f"AND `{pk2}` IN {df[pk2].unique().to_list()}"
            else:
                raise TypeError(df.dtypes[idx2])
        with self.table_locks[table_name]:
            rs: List[Tuple] = self.execute(sql, columnar=True)
            if rs:
                # 多列反选
                exists: pl.DataFrame = pl.DataFrame(dict(zip(pks, rs)))
                converts = []
                for pk in pks:
                    if exists[pk].dtype != df[pk].dtype:
                        converts.append(pl.col(pk).cast(df[pk].dtype))
                if converts:
                    exists = exists.with_columns(converts)
                df = df.join(exists, on=pks, how='anti')
            # 确定df转换为csv时候的时间格式
            return self.import_csv(
                table_name=table_name,
                csv=df.write_csv(include_header=True, datetime_format='%Y-%m-%d %H:%M:%S%.9f'),
                height=df.height
            )



# -*- coding: utf-8 -*-
import enum
from enum import EnumMeta, Enum
from typing import Any, Dict


class ByKey(EnumMeta):
    map: Dict[str, Any] = {}

    def __getattribute__(cls, name: str):
        value = super().__getattribute__(name)
        if isinstance(value, cls):
            value = cls.map[name]
        return value

    @classmethod
    def check(cls, obj: enum.EnumMeta):
        """一致性检查"""
        errors: str = ",".join(sorted(set(cls.map) - {c.name for c in obj}))  # type: ignore
        if errors:
            raise ValueError(f"{obj.__class__.__name__}缺少枚举项: {errors}")
        errors: str = ",".join(sorted({c.name for c in obj} - set(cls.map)))  # type: ignore
        if errors:
            raise ValueError(f"{obj.__class__.__name__}非法枚举项: {errors}")



# -*- coding: utf-8 -*-
import asyncio
import functools
from loguru import logger


def aioretry(
        max_retries: int = 1,
        retry_interval: float = 1.0,
        log: bool = True
):
    """异步重试装饰器"""

    def decorator(func):
        @functools.wraps(func)
        async def decorated(*args, **kwargs):
            for retry in range(max_retries):
                try:
                    return await func(*args, **kwargs)
                except Exception as e:
                    if retry == max_retries - 1:
                        if log:
                            logger.error(f'{max_retries}次重试失败: {args} {kwargs}')
                        raise e
                    else:
                        if log:
                            logger.error(f'[{e.__class__.__name__}]{e}')
                        await asyncio.sleep(retry_interval)

        return decorated

    return decorator



# -*- coding: utf-8 -*-
import datetime
import decimal
import inspect
import lz4.frame
import math
import os
import pickle
import re
import sqlite3
import threading
import time
from typing import Any, Set, List, Tuple, Dict
from .lock import SyncRWLock


class Cache:
    """持久化TTL缓存"""

    def __init__(self, path: str = None):
        current_frame = inspect.currentframe()
        caller_frame = current_frame.f_back
        caller_file = caller_frame.f_code.co_filename
        if not path:
            path: str = f'_{os.path.splitext(os.path.basename(caller_file))[0]}.db3'
        self.path: str = os.path.join(os.path.dirname(caller_file), path)
        self.default_dataset: str = 'DEFAULT'
        self.compress_min_size: int = max(64, len(lz4.frame.compress(b'\x00')))
        self.conn = sqlite3.connect(self.path, check_same_thread=False)
        self.lock = SyncRWLock(reentrant=True)
        # self.conn.execute('PRAGMA auto_vacuum = FULL')
        self.datasets: Set[str] = self.get_tables()
        self.upgrade_tables()

    def get_tables(self) -> Set[str]:
        """获取所有数据集"""
        with self.lock:
            cursor = self.conn.cursor()
            r = cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
            return {row[0] for row in r.fetchall()}

    def upgrade_tables(self):
        """升级表到最新版本"""
        self.lock.write_acquire()
        try:
            updated: bool = False
            for dataset in self.datasets:
                cursor = self.conn.cursor()
                r = cursor.execute(f"PRAGMA table_info({dataset});")
                cols: List[str] = [x[1] for x in r.fetchall()]
                if 'encoding' not in cols:
                    self.conn.execute(f'ALTER TABLE `{dataset}` ADD COLUMN `encoding` INTEGER;')
                    self.conn.execute(f'UPDATE `{dataset}` SET `encoding` = 5;')
                    updated |= True
            if updated:
                self.conn.commit()
        finally:
            self.lock.write_release()

    def create_table(self, dataset: str) -> None:
        """创建数据集"""
        self.lock.write_acquire()
        try:
            assert not re.findall(r'\s', dataset)
            self.conn.execute(f"""CREATE TABLE IF NOT EXISTS `{dataset}` (
                `key` TEXT PRIMARY KEY NOT NULL,
                `value` BLOB,
                `encoding` INTEGER NOT NULL,
                `expired_at` REAL NOT NULL
            )""")
            self.conn.execute(f"""CREATE INDEX IF NOT EXISTS expired_at ON `{dataset}` (`expired_at`)""")
            self.conn.commit()
        finally:
            self.lock.write_release()

    def encode_value(self, value: Any) -> Tuple[Any, int]:
        """数据编码"""
        if value is None:
            return value, 0     # 原始数据
        elif isinstance(value, bool):
            return value, 64
        elif isinstance(value, datetime.datetime):
            return value, 65
        elif isinstance(value, datetime.date):
            return value, 66
        elif isinstance(value, decimal.Decimal):
            return str(value), 67
        elif isinstance(value, int):
            if - 2 ** 63 <= value <= 2 ** 63 - 1:
                return value, 0  # 原始数据
            else:
                return str(value), 68
        elif isinstance(value, float):
            return value, 0     # 原始数据
        elif isinstance(value, str):
            value: bytes = value.encode('utf-8')
            if len(value) <= self.compress_min_size:
                return value, 1     # 字符串utf-8编码
            else:
                return lz4.frame.compress(value), 2     # 字符串utf-8编码并压缩
        elif isinstance(value, bytes):
            if len(value) <= self.compress_min_size:
                return value, 0     # 原始数据
            else:
                return lz4.frame.compress(value), 3     # 仅压缩
        else:
            payload: bytes = pickle.dumps(value)
            if len(payload) <= self.compress_min_size:
                return payload, 4                         # 仅序列化
            else:
                return lz4.frame.compress(payload), 5     # 序列化并压缩

    @classmethod
    def decode_value(cls, value: Any, encoding: int) -> Any:
        """数据解码"""
        match encoding:
            case 0:
                return value
            case 1:
                return value.decode('utf-8')
            case 2:
                return lz4.frame.decompress(value).decode('utf-8')
            case 3:
                return lz4.frame.decompress(value)
            case 4:
                return pickle.loads(value)
            case 5:
                return pickle.loads(lz4.frame.decompress(value))
            case 64:
                return bool(value)
            case 65:
                return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S.%f')
            case 66:
                return datetime.datetime.strptime(value, '%Y-%m-%d').date()
            case 67:
                return decimal.Decimal(value)
            case 68:
                return int(value)
            case _:
                raise ValueError(encoding)

    def set(self, dataset: str = None, key: str = '', value: Any = None, expire: float = math.inf) -> None:
        """设置键值"""
        self.lock.write_acquire()
        try:
            if not dataset:
                dataset = self.default_dataset
            assert isinstance(dataset, str), 'dataset必须是字符串'
            assert dataset, 'dataset不能为空'
            assert isinstance(key, str), 'key必须是字符串'
            assert key, 'key不能为空'
            if dataset not in self.datasets:
                self.create_table(dataset=dataset)
                self.datasets.clear()
                self.datasets.update(self.get_tables())
            payload, encoding = self.encode_value(value=value)
            sql: str = f"""REPLACE INTO `{dataset}` (`key`, `value`, `encoding`, `expired_at`) VALUES(?,?,?,?)"""
            self.conn.execute(sql, (key, payload, encoding, time.time() + expire))
            self.conn.commit()
        finally:
            self.lock.write_release()

    def rename(self, dataset: str = None, key: str = '', new_key: str = '') -> None:
        """重命名键值"""
        self.lock.write_acquire()
        try:
            if not dataset:
                dataset = self.default_dataset
            assert isinstance(dataset, str), 'dataset必须是字符串'
            assert dataset, 'dataset不能为空'
            assert isinstance(key, str), 'key必须是字符串'
            assert key, 'key不能为空'
            assert isinstance(new_key, str), 'new_key必须是字符串'
            assert new_key, 'new_key不能为空'
            if key == new_key:
                return
            if dataset not in self.datasets:
                self.create_table(dataset=dataset)
                self.datasets.clear()
                self.datasets.update(self.get_tables())
            sql: str = f"""UPDATE `{dataset}` SET `key` = ? WHERE `key` = ?"""
            self.conn.execute(sql, (new_key, key))
            self.conn.commit()
        finally:
            self.lock.write_release()

    def update(self, dataset: str = None, items: Dict[str, Any] = None, expire: float = math.inf) -> None:
        """批量设置键值"""
        self.lock.write_acquire()
        try:
            if not dataset:
                dataset = self.default_dataset
            assert isinstance(dataset, str), 'dataset必须是字符串'
            assert dataset, 'dataset不能为空'
            assert isinstance(items, dict), 'items必须是字典'
            if not items:
                return
            for key in items.keys():
                assert isinstance(key, str), 'key必须是字符串'
                assert key, 'key不能为空'
            if dataset not in self.datasets:
                self.create_table(dataset=dataset)
                self.datasets.clear()
                self.datasets.update(self.get_tables())
            ps: List[Tuple] = []
            now = time.time()
            for key, value in items.items():
                payload, encoding = self.encode_value(value=value)
                ps.append((key, payload, encoding, now + expire))
            sql: str = f"""REPLACE INTO `{dataset}` (`key`, `value`, `encoding`, `expired_at`) VALUES(?,?,?,?)"""
            self.conn.executemany(sql, ps)
            self.conn.commit()
        finally:
            self.lock.write_release()

    def add(self, dataset: str = None, key: str = '', value: Any = None, expire: float = math.inf) -> None:
        """添加值到集合中"""
        self.lock.write_acquire()
        try:
            if not dataset:
                dataset = self.default_dataset
            assert isinstance(dataset, str), 'dataset必须是字符串'
            assert dataset, 'dataset不能为空'
            assert isinstance(key, str), 'key必须是字符串'
            assert key, 'key不能为空'
            rs: Set = self.get(dataset=dataset, key=key) or set()
            try:
                hash(value)
                rs.add(value)
            except TypeError:
                rs.update(value)
            self.set(dataset=dataset, key=key, value=rs, expire=expire)
        finally:
            self.lock.write_release()

    def delete(self, dataset: str = None, key: str | List[str] = None) -> None:
        """删除键值对"""
        self.lock.write_acquire()
        try:
            if not dataset:
                dataset = self.default_dataset
            assert isinstance(dataset, str), 'dataset必须是字符串'
            assert dataset, 'dataset不能为空'
            assert key, 'key不能为空'
            if isinstance(key, str):
                key: List[str] = [key]
            elif isinstance(key, list):
                for k in key:
                    assert isinstance(k, str), 'key列表元素必须是字符串'
                    assert k, 'key列表元素不能为空'
            else:
                raise ValueError('key必须是字符串或字符串列表')
            if dataset in self.datasets:
                placeholders: str = ', '.join(['?'] * len(key))
                sql: str = f"""DELETE FROM `{dataset}` WHERE `key` IN ({placeholders})"""
                self.conn.execute(sql, key)
                self.conn.commit()
        finally:
            self.lock.write_release()

    def get(self, dataset: str = None, key: str = '') -> Any | None:
        """获取键值"""
        r = None
        with self.lock:
            if not dataset:
                dataset = self.default_dataset
            assert isinstance(dataset, str), 'dataset必须是字符串'
            assert dataset, 'dataset不能为空'
            assert isinstance(key, str), 'key必须是字符串'
            assert key, 'key不能为空'
            if dataset in self.datasets:
                cursor = self.conn.cursor()
                sql: str = f"""SELECT `value`, `encoding` FROM `{dataset}` WHERE `key` = ? AND `expired_at` > ?"""
                r = cursor.execute(sql, (key, time.time()))
                r = r.fetchone()
        if r is None:
            return r
        else:
            value, encoding = r
            return self.decode_value(value=value, encoding=encoding)

    def keys(self, dataset: str = None) -> List[str] | None:
        """列出所有键名"""
        with self.lock:
            if not dataset:
                dataset = self.default_dataset
            assert isinstance(dataset, str), 'dataset必须是字符串'
            assert dataset, 'dataset不能为空'
            if dataset in self.datasets:
                cursor = self.conn.cursor()
                sql: str = f"""SELECT `key` FROM `{dataset}` WHERE `expired_at` > ?"""
                r = cursor.execute(sql, (time.time(), ))
                result: List[str] = []
                for (key, ) in r.fetchall():
                    result.append(key)
                return result

    def items(self, dataset: str = None) -> Dict:
        """列出所有键值对"""
        rs: List[Tuple] = []
        result: Dict[str, Any] = {}
        with self.lock:
            if not dataset:
                dataset = self.default_dataset
            assert isinstance(dataset, str), 'dataset必须是字符串'
            assert dataset, 'dataset不能为空'
            if dataset in self.datasets:
                cursor = self.conn.cursor()
                sql: str = f"""SELECT `key`, `value`, `encoding` FROM `{dataset}` WHERE `expired_at` > ?"""
                r = cursor.execute(sql, (time.time(), ))
                rs = r.fetchall()
        for key, value, encoding in rs:
            result[key] = self.decode_value(value=value, encoding=encoding)
        return result

    def drop_table(self, dataset: str):
        """删除数据集"""
        self.lock.write_acquire()
        try:
            assert isinstance(dataset, str), 'dataset必须是字符串'
            assert dataset, 'dataset不能为空'
            if dataset in self.datasets:
                sql: str = f"""DROP TABLE IF EXISTS `{dataset}`"""
                self.conn.execute(sql)
                self.conn.commit()
                self.datasets.discard(dataset)
        finally:
            self.lock.write_release()

    def vacuum(self, force: bool = True):
        self.lock.write_acquire()
        try:
            key_vacuum_at: str = '__VACUUM__'
            vacuum_at: int | None = self.get(key=key_vacuum_at)
            if force or not vacuum_at or time.time() - vacuum_at >= 86400:
                self.conn.execute('VACUUM')
                self.set(key=key_vacuum_at, value=int(time.time()))
        finally:
            self.lock.write_release()

    def close(self):
        self.lock.write_acquire()
        try:
            self.conn.commit()
            for dataset in self.datasets:
                sql: str = f"DELETE FROM `{dataset}` WHERE `expired_at` <= ?"
                self.conn.execute(sql, (int(time.time()), ))
            self.vacuum(force=False)
            self.conn.close()
        finally:
            self.lock.write_release()



# -*- coding: utf-8 -*-
import datetime
import functools
import math
import statistics
from collections import OrderedDict
from typing import List, Tuple, Dict


class NavIndexes:
    """基金净值指标"""

    def __init__(self, values: Dict[datetime.date, float], r: float):
        """
        :param values: 每日复权净值序列
        :param r: 无风险利率
        """
        self.r: float = r
        assert len(values) >= 2
        self.start_date: datetime.date = min(values)
        self.end_date: datetime.date = max(values)
        # 检查完整性
        self.values: OrderedDict[datetime.date, float] = OrderedDict()
        for date, value in sorted(values.items()):
            assert isinstance(value, (int, float))
            self.values[date] = value
            assert value > 0
        start_year = datetime.date(year=self.start_date.year, month=1, day=1)
        end_year = datetime.date(year=self.end_date.year, month=12, day=31)
        # 平均每年天数,介于365与366之间
        self.annualized_days: float = (end_year - start_year).days / (self.end_date.year - self.start_date.year + 1)
        # 每年交易日天数,即每年有净值的天数,通常在250天左右
        self.annualized_trade_days: float = (
            (len(values) - 1) / (self.end_date - self.start_date).days * self.annualized_days
        )

    def calculate_indexs(self):
        """综合指标计算"""
        mdd, mdd_peak_date, mdd_date, mdd_recover_date = self.max_draw_down  # 累计收益率
        return {
            "mdd": mdd,  # 最大回撤比
            "mdd_peak_date": mdd_peak_date,  # 最大回撤起始日期
            "mdd_date": mdd_date,  # 最大回撤终止日期
            "mdd_recover_date": mdd_recover_date,  # 最大回撤修复日期
            "accumulated_return": self.accumulated_return,  # 累计收益率
            "annualized_return": self.annualized_return,  # 年化收益率
            "calmar_ratio": self.calmar_ratio,  # 卡玛比率
            "sharpe_ratio": self.sharpe_ratio,  # 夏普比率
            "sortino_ratio": self.sortino_ratio,  # 索提诺比率
            "average_daily_return": self.average_daily_return,  # 日均收益均值
            "stdev": self.stdev,  # 标准差
            "annualized_stdev": self.annualized_stdev,  # 年化标准差
            "downside_stdev": self.downside_stdev,  # 下行标准差
        }

    @property
    @functools.lru_cache()
    def max_draw_down(self) -> Tuple[float, datetime.date, datetime.date, datetime.date]:
        """
        计算最大回撤及修复时间
        :rtype: 最大回撤比, 最大回撤起始日期, 最大回撤终止日期, 最大回撤修复日期
        """
        mdd_date: datetime.date = self.start_date
        mdd: float = 0.0
        mdd_peak_date: datetime.date = self.start_date
        scan_peak_date: datetime.date = self.start_date
        scan_peak: float = 0.0
        mdd_recover_date: datetime.date = self.start_date
        for date, value in self.values.items():
            if value > scan_peak:
                scan_peak_date = date
                scan_peak = value
            dd: float = 0.0 if scan_peak == 0 else 1 - value / scan_peak
            if dd > mdd:
                # 发生最大回撤
                mdd_date = date
                mdd = dd
                mdd_peak_date = scan_peak_date
                mdd_recover_date = self.start_date
            elif dd <= 0 and mdd_recover_date == self.start_date:
                mdd_recover_date = date
        return mdd, mdd_peak_date, mdd_date, mdd_recover_date

    @property
    @functools.lru_cache()
    def accumulated_return(self) -> float:
        """累计收益率"""
        v1: float = self.values[self.start_date]
        v2: float = self.values[self.end_date]
        return v2 / v1 - 1

    @property
    @functools.lru_cache()
    def annualized_return(self) -> float:
        """年化收益率"""
        return self.accumulated_return / (self.end_date - self.start_date).days * self.annualized_days

    @property
    @functools.lru_cache()
    def calmar_ratio(self) -> float:
        """卡玛比率"""
        return self.annualized_return / self.max_draw_down[0]

    @property
    @functools.lru_cache()
    def sharpe_ratio(self):
        """夏普比率"""
        return (self.annualized_return - self.r) / self.annualized_stdev

    @property
    @functools.lru_cache()
    def sortino_ratio(self) -> float:
        """索提诺比率"""
        return (self.annualized_return - self.r) / self.downside_stdev

    @property
    @functools.lru_cache()
    def daily_return(self) -> OrderedDict[datetime.date, float]:
        """日均收益"""
        dr: OrderedDict[datetime.date, float] = OrderedDict()
        last_value: float = self.values[self.start_date]
        for i, (date, value) in enumerate(self.values.items()):
            if i == 0:
                dr[date] = 0.0
            else:
                dr[date] = 1 - last_value / value
                last_value = value
        return dr

    @property
    @functools.lru_cache()
    def average_daily_return(self) -> float:
        """日均收益均值"""
        return statistics.mean(self.daily_return.values())

    @property
    @functools.lru_cache()
    def stdev(self) -> float:
        """标准差"""
        return statistics.stdev(self.daily_return.values())

    @property
    @functools.lru_cache()
    def annualized_stdev(self) -> float:
        """年化标准差"""
        return self.stdev * math.sqrt(self.annualized_trade_days)

    @property
    @functools.lru_cache()
    def downside_stdev(self) -> float:
        """下行标准差"""
        values: List[float] = [value for value in self.daily_return.values() if value < self.r]
        if len(values) >= 2:
            return statistics.stdev(values)
        else:
            return 0.0



# -*- coding: utf-8 -*-
import asyncio
import json
from typing import Any, Callable, Union
from loguru import logger
import websockets
import websockets.client


class AuthClient:
    def __init__(
        self,
        uri: str,
        generator: Callable,
        connect_callback: Union[Callable, None] = None,
        event_callback: Union[Callable, None] = None,
    ):
        assert uri and generator
        self.uri: str = uri
        self.generator: Callable = generator
        self.connect_callback: Union[Callable, None] = connect_callback
        self.event_callback: Union[Callable, None] = event_callback
        self.ws: websockets.client.WebSocketClientProtocol | None = None
        self.ready: asyncio.Event = asyncio.Event()
        self.app_id: str = ""
        self.conn_id: int = -1
        self.req_id: int = -1
        self.responses: dict[str, asyncio.Queue] = {}

        self.connect_task: asyncio.Task = asyncio.create_task(self.connect_ws())

    @property
    def request_id(self):
        self.req_id += 1
        return f"{self.conn_id}_{self.req_id}"

    async def connect_ws(self):
        while not self.ready.is_set():
            set_app_task: asyncio.Task | None = None
            ping_task: asyncio.Task | None = None
            try:
                logger.debug(f"connecting to AuthServer...")
                self.ws = await websockets.connect(  # type: ignore
                    uri=self.uri,
                    max_size=2**32 - 1,
                    max_queue=2**8,
                    read_limit=2**32 - 1,
                    write_limit=2**32 - 1,
                    ssl=None,
                )
                logger.success(f"connected to AuthServer!")
                self.ready.set()
                if self.connect_callback:
                    await self.connect_callback(connected=True)
                while self.ready:
                    payload: dict | None = await self.recv()
                    if not payload:
                        continue
                    response_id: str | None = payload.get("response_id", None)
                    if response_id:
                        await self.responses[response_id].put(payload)
                    if payload.get("event") == "push_connection_id":
                        self.conn_id = payload.get("data").get("connection_id")  # type: ignore
                        self.req_id = -1
                        set_app_task = asyncio.create_task(self.set_app())
                        ping_task = asyncio.create_task(self.start_ping())
                        continue
                    if self.event_callback:
                        await self.event_callback(payload)
            except Exception as e:
                logger.error(f"[{e.__class__.__name__}]{e}")
            finally:
                logger.warning(f"AuthServer disconnected")
                # 取消任务
                if ping_task:
                    ping_task.cancel()
                if set_app_task:
                    set_app_task.cancel()
                # 关闭连接
                if self.ws:
                    await self.ws.close()
                    self.ws = None
                self.ready.clear()
                self.app_id = ""
                if self.connect_callback:
                    await self.connect_callback(connected=False)
                await asyncio.sleep(3)
        # 取消重连
        self.connect_task.cancel()

    async def wait_response(self, request_id: str) -> tuple[Any, str]:
        error: str = ""
        data: Any = None
        try:
            rj: dict = await asyncio.wait_for(
                self.responses[request_id].get(), timeout=10
            )
            if rj.get("code") != 0:
                error = rj["message"]
            data: Any = rj.get("data")
        except Exception as e:
            error = e.__class__.__name__
        return error, data

    async def send(self, data: dict):
        try:
            if self.ws:
                message: str = json.dumps(data)
                # logger.info(f" send {message}")
                await self.ws.send(message)
                self.responses[data["request_id"]] = asyncio.Queue()
        except Exception as e:
            logger.error(f"[{e.__class__.__name__}]{e}")
            if self.ws:
                await self.ws.close()
                self.ws = None

    async def recv(self) -> dict | None:
        payload: str = await self.ws.recv()  # type: ignore
        try:
            if payload:
                # logger.debug(f"recv {payload}")
                return json.loads(payload)
        except Exception as e:
            logger.error(f"[{e.__class__.__name__}]{e}")
        return None

    async def login(self, info: dict) -> tuple[str, Any]:
        if not self.ready.is_set():
            return "鉴权模块未连接", None
        request_id: str = self.request_id
        data: dict = {"event": "login", "request_id": request_id, "data": info}
        await self.send(data)
        error, response = await self.wait_response(request_id)
        return error, response

    async def logout_by_session_id(self, info: dict) -> tuple[str, Any]:
        if not self.ready.is_set():
            return "鉴权模块未连接", None
        request_id: str = self.request_id
        data: dict = {
            "event": "logout_by_session_id",
            "request_id": request_id,
            "data": info,
        }
        await self.send(data)
        error, response = await self.wait_response(request_id)
        return error, response

    async def set_app(self):
        while not self.app_id:
            await self.ready.wait()
            request_id: str = self.request_id
            send_data: dict = {
                "request_id": request_id,
                "event": "set_app",
                "data": self.generator(),
            }
            await self.send(send_data)
            try:
                response: dict = await asyncio.wait_for(
                    self.responses[request_id].get(), timeout=3
                )
                code: int | None = response.get("code")
                data: dict | None = response.get("data")
                if code == 0 and data:
                    self.app_id = data.get("app_id")
                    logger.success(f"set app success,appid={self.app_id}")
            except asyncio.TimeoutError as e:
                logger.error(f"[{e.__class__.__name__}]{e}")
            except Exception as e:
                logger.error(f"[{e.__class__.__name__}]{e}")
            await asyncio.sleep(1)

    async def start_ping(self, interval=3):
        while True:
            await self.ready.wait()
            request_id: str = self.request_id
            data: dict = {
                "request_id": request_id,
                "event": "ping",
            }
            try:
                await self.send(data)
                await asyncio.wait_for(self.responses[request_id].get(), timeout=3)
            except asyncio.TimeoutError as e:
                logger.error(f"[{e.__class__.__name__}]{e}")
                await self.ws.close(reason="ping超时")  # type: ignore
            await asyncio.sleep(interval)



# -*- coding: utf-8 -*-
"""外汇基础货币枚举"""
import enum


class Currency(enum.Enum):
    AUD = enum.auto()
    BRL = enum.auto()
    CAD = enum.auto()
    CHF = enum.auto()
    CNH = enum.auto()
    CZK = enum.auto()
    DKK = enum.auto()
    EUR = enum.auto()
    GBP = enum.auto()
    HKD = enum.auto()
    JPY = enum.auto()
    MXN = enum.auto()
    NOK = enum.auto()
    NZD = enum.auto()
    PLN = enum.auto()
    RUB = enum.auto()
    SEK = enum.auto()
    SGD = enum.auto()
    TRY = enum.auto()
    USD = enum.auto()
    XAG = enum.auto()
    XAU = enum.auto()
    XPD = enum.auto()
    XPT = enum.auto()
    ZAR = enum.auto()


class Forex:
    def __init__(self, symbol: str):
        assert len(symbol) == 6
        self.currency1: Currency = Currency[symbol[:3].upper()]
        self.currency2: Currency = Currency[symbol[3:].upper()]
        self.symbol: str = f'{self.currency1.name}{self.currency2.name}'

    def __repr__(self):
        return self.symbol



# -*- coding: utf-8 -*-
"""东方财富代码"""
import datetime
import pypinyin
import re
import string
from .exchange import Exchange
from .symbol import (
    SymbolType, SymbolSubType, StockType, FundType, BondType, FuturesType,
    OptionsType, WarrantsType, IndexType, DrType, SpotType, SymbolFlag)
from typing import Any, AnyStr, Dict, List, Set, Tuple


Trans: Dict[int, int] = {ord(char) + 0xFEE0: ord(char) for char in string.digits + string.ascii_letters}
EMMarkets = [
    0, 1, 2, 8, 10, 11, 12, 46, 47, 50, 90,
    100, 101, 102, 103, 104, 105, 106, 107, 108, 109,
    110, 111, 112, 113, 114, 115, 116, 118, 119,
    120, 121, 122, 123, 124, 125, 128,
    130, 131, 132, 133, 134, 139,
    140, 141, 142, 144,
    150, 151, 153, 154, 155, 156, 157, 158,
    159,    # 商品加权指数
    160, 162, 163, 165,
    171,
    200, 201, 202,
    220, 221, 223, 225, 226,
    245, 246, 248,
    251, 252,
    302, 303, 304, 305, 306, 307, 308, 309,
    310, 311, 312, 313, 315, 316, 318, 319,
    320, 321, 322, 323, 324, 325, 326, 327, 329,
    330, 331, 332, 333, 334, 335, 336, 337, 338, 339,
    340, 341, 342, 343, 344, 345, 346, 347, 348, 349,
    351, 352, 356,
]


class KLineRecord:
    """K线数据类"""
    def __init__(self, kline: AnyStr, fields2: List[str]):
        r: Dict = dict(zip(fields2, kline.split(',')))
        self._datetime: datetime.datetime = datetime.datetime.strptime(r['f51'], '%Y-%m-%d')
        self.date: str = self._datetime.strftime('%Y-%m-%d')
        self.open: float = float(r['f52'])
        self.close: float = float(r['f53'])
        self.high: float = float(r['f54'])
        self.low: float = float(r['f55'])
        self.volume: float = float(r['f56'])
        self.money: float = float(r['f57'])
        self.turnover_rate: float = float(r['f61'])
        self.open_interest: float = float(r['f63'])
        self.settlement_price: float = float(r['f65'])

    @property
    def raw_datetime(self) -> datetime.datetime:
        return self._datetime

    def to_list(self) -> List[Any]:
        return list([v for k, v in self.__dict__.items() if not k.startswith('_')])

    @property
    def yyyymmdd(self) -> str:
        return self._datetime.strftime('%Y%m%d')


class TrendRecord:
    """分时线数据类"""
    def __init__(self, trend: AnyStr, fields2: List[str]):
        r: Dict = dict(zip(fields2, trend.split(',')))
        self._datetime: datetime.datetime = datetime.datetime.strptime(r['f51'], '%Y-%m-%d %H:%M')
        self.datetime: str = self._datetime.strftime('%Y-%m-%d %H:%M:%S')
        self.open: float = float(r['f52'])
        self.close: float = float(r['f53'])
        self.high: float = float(r['f54'])
        self.low: float = float(r['f55'])
        self.volume: float = float(r['f56'])
        self.money: float = float(r['f57'])
        self.open_interest: float = float(r['f59'])

    @property
    def raw_datetime(self) -> datetime.datetime:
        return self._datetime

    @property
    def date(self) -> str:
        return self._datetime.strftime('%Y-%m-%d')

    @property
    def time(self) -> str:
        return self._datetime.strftime('%H%M%S')

    def to_list(self) -> List[Any]:
        return list([v for k, v in self.__dict__.items() if not k.startswith('_')])

    @property
    def yyyymmdd(self) -> str:
        return self._datetime.strftime('%Y%m%d')


class EMParser:
    fmt_date: str = '%Y-%m-%d'
    fmt_time: str = '%H:%M:%S'
    fmt_datetime: str = f'{fmt_date} {fmt_time}'

    def __init__(self, record: Dict):
        self.record: Dict = record

    def strip(self, fn: str) -> Any:
        value: Any = self.record[fn]
        if isinstance(value, str) and value == '-':
            return
        return value

    def parse_base_info(self) -> Tuple[Exchange, SymbolSubType, SymbolFlag]:
        """从东方财富市场信息解析交易所及标的品种类型等信息"""
        m = self.m
        t = self.t
        s = self.s
        e = self.e
        f = self.f
        exchange: Exchange | None = None
        typ: SymbolSubType | None = None
        flg: SymbolFlag = SymbolFlag()

        if m in {0, 1}:
            if f & 2 ** 2:
                flg.is_st = True
            if f & 2 ** 3:
                flg.is_new = True
            if s & 2 ** 19:
                flg.is_approval = True
            if s & 2 ** 17:
                flg.is_registration = True
            if s & 2 ** 9:
                flg.is_innovation_market = True
            if s & 2 ** 8:
                flg.is_base_market = True
            if s & 2 ** 7:
                flg.is_call_auction = True
            if s & 2 ** 5:
                flg.is_market_making = True
            if s & 3:
                flg.is_pt = True

        if m == 0:
            exchange = Exchange.SZSE
            if t == 5:
                typ = IndexType.INDEX_OTHER
            elif t == 6:
                typ = StockType.STOCK_A
            elif t == 7:
                typ = StockType.STOCK_B
            elif t == 8:
                # [7, 9, 10, 11, 12, 13, 14, 15, 29]
                if e == 7:
                    typ = BondType.BOND_OTHER         # 政府债
                elif e == 9:
                    typ = BondType.BOND_ENTERPRISE
                elif e == 10:
                    typ = BondType.BOND_COMPANY
                elif e == 11:
                    typ = BondType.BOND_CONVERTIBLE
                elif e == 12:
                    typ = BondType.BOND_EXCHANGEABLE
                elif e == 13:
                    typ = BondType.BOND_ABS
                elif e == 14:
                    typ = BondType.BOND_BUYBACK
                elif e == 15:
                    typ = BondType.BOND_COMPANY       # 私募公司债
                elif e == 29:
                    typ = StockType.STOCK_P
            elif t == 10:
                if e == 16:
                    typ = FundType.FUND_ETF
                elif e == 97:
                    typ = FundType.FUND_REITS
            elif t == 80:
                typ = StockType.STOCK_G
            elif t == 81:
                if s & 2048:
                    # 北证A股
                    exchange = Exchange.BSE
                    typ = StockType.STOCK_A
                else:
                    exchange = Exchange.NEEQ
                    typ = StockType.STOCK_OTHER
        elif m == 1:
            exchange = Exchange.SSE
            # if e == 15:
            #     return exchange, stype  # 新标准券
            # elif e == 34:
            #     return exchange, stype  # 申购
            # elif e == 36:
            #     return exchange, stype  # 配号
            # elif e == 37:
            #     return exchange, stype  # 认购款
            # elif e == 44:
            #     return exchange, stype  # 配债
            # elif e == 48:
            #     return exchange, stype  # 分红代码
            # elif e == 49:
            #     return exchange, stype  # 转托管代码
            if t == 1:
                typ = IndexType.INDEX_OTHER
            elif t == 2:
                typ = StockType.STOCK_A
            elif t == 3:
                typ = StockType.STOCK_B
            elif t == 4:
                # [7, 9, 10, 11, 12, 14, 15, 29, 34, 36, 37, 44, 99]
                if e == 7:
                    typ = BondType.BOND_OTHER         # 政府债
                elif e == 9:
                    typ = BondType.BOND_ENTERPRISE    # 一般企业债
                elif e == 10:
                    typ = BondType.BOND_COMPANY       # 一般公司债
                elif e == 11:
                    typ = BondType.BOND_CONVERTIBLE   # 可转债
                elif e == 12:
                    typ = BondType.BOND_EXCHANGEABLE
                elif e == 14:
                    typ = BondType.BOND_BUYBACK
                elif e == 29:
                    typ = StockType.STOCK_P
                elif e == 99:
                    pass    # 定向可转债 属于私募债
            elif t == 9:
                # [16, 17, 20, 23, 34, 37, 46, 47, 48, 49, 97]
                if e == 16:
                    typ = FundType.FUND_ETF
                elif e == 17:
                    typ = FundType.FUND_LOF
                elif e == 20:
                    typ = FundType.FUND_BOND
                elif e == 23:
                    typ = FundType.FUND_HYBRID
                elif e == 46:
                    typ = FundType.FUND_OPEN    # 开放式
                elif e == 47:
                    pass
                elif e == 97:
                    typ = FundType.FUND_REITS
            elif t == 23:
                typ = StockType.STOCK_K
        elif m == 2:
            exchange = Exchange.CSI
            if t == 24:
                typ = IndexType.INDEX_OTHER
        elif m == 10:
            exchange = Exchange.SSE
            typ = OptionsType.OPTIONS_ETF
            if t == 173:
                flg.is_call = True
            elif t == 174:
                flg.is_put = True
        elif m == 12:
            exchange = Exchange.SZSE
            typ = OptionsType.OPTIONS_ETF
            if t == 178:
                flg.is_call = True
            elif t == 179:
                flg.is_put = True
        elif m == 128:
            exchange = Exchange.HKEX
            if t == 1:
                typ = StockType.STOCK_A   # 港股主板
            elif t == 2:
                typ = StockType.STOCK_A  # 港股主板
            elif t == 3:
                typ = StockType.STOCK_A   # 港股主板
            elif t == 4:
                typ = StockType.STOCK_G   # 港股创业板
            elif t == 5:
                typ = WarrantsType.WARRANTS_CBBC  # 港股牛熊证
            elif t == 6:
                typ = OptionsType.OPTIONS_STOCK   # 港股窝轮
            if s & 64:
                pass    # 人民币交易港股
            if s & 1:
                pass    # ADR
        if m in {124, 125, 305}:
            exchange = Exchange.HKEX
            typ = IndexType.INDEX_OTHER
        elif m == 101:
            exchange = Exchange.COMEX
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 102:
            exchange = Exchange.NYMEX
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 103:
            exchange = Exchange.CBOT
            if t in {12, 20}:
                typ = FuturesType.FUTURES_INDEX
            elif t in {13, 14, 15, 16, 17}:
                typ = FuturesType.FUTURES_BOND
            else:
                typ = FuturesType.FUTURES_COMMODITY
        elif m == 104:
            exchange = Exchange.SGX
            if t == 1:
                typ = FuturesType.FUTURES_INDEX
            else:
                typ = FuturesType.FUTURES_COMMODITY
        elif m == 105:
            exchange = Exchange.NASDAQ
            typ = StockType.STOCK_A
        elif m == 106:
            exchange = Exchange.NYSE
            typ = StockType.STOCK_A
        elif m == 107:
            exchange = Exchange.AMEX
            typ = StockType.STOCK_A
        elif m == 108:
            exchange = Exchange.NYBOT
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 109:
            exchange = Exchange.LME
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 110:
            exchange = Exchange.BMD
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 111:
            exchange = Exchange.TOCOM
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 112:
            exchange = Exchange.ICE
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 113:
            exchange = Exchange.SHFE
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 114:
            exchange = Exchange.DCE
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 115:
            exchange = Exchange.ZCE
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 116:
            # t = [0, 1, 2, 3, 4, 5, 6, 7]
            # e = [9, 10, 15]
            exchange = Exchange.HKEX
            if t == 0:
                pass
            elif t == 1:
                # e = [0, 8, 9, 10, 16, 17, 18]
                if e & 0:
                    typ = FundType.FUND_ETF
                elif e & 8:
                    typ = FundType.FUND_REITS
                elif e & 9:
                    typ = FundType.FUND_OTHER
                elif e & 10:
                    typ = FundType.FUND_ETF
                elif e & 16:
                    typ = FundType.FUND_ETF
                elif e & 17:
                    typ = FundType.FUND_BOND
                elif e & 18:
                    typ = FundType.FUND_DERIVATIVES
            elif t == 2:
                if e & 4:
                    typ = BondType.BOND_OTHER
            elif t == 3:
                # e = [0, 1, 2, 7]
                typ = StockType.STOCK_A
                if e & 2:
                    typ = StockType.STOCK_P
            elif t == 4:
                typ = StockType.STOCK_G
            elif t == 5:
                typ = WarrantsType.WARRANTS_CBBC
                if s & 2:
                    flg.is_call = True        # 牛证
                if s & 4:
                    flg.is_put = True         # 熊证
            elif t == 6:
                typ = OptionsType.OPTIONS_STOCK
                if s & 2:
                    flg.is_call = True        # 购
                if s & 4:
                    flg.is_put = True         # 沽
            elif t == 7:
                typ = WarrantsType.WARRANTS_INLINE
        elif m == 118:
            exchange = Exchange.SGE
            typ = SpotType.SPOT_COMMODITY
        elif m == 139:
            exchange = Exchange.HKEX
            typ = OptionsType.OPTIONS_FOREX
        elif m == 140:
            exchange = Exchange.DCE
            typ = OptionsType.OPTIONS_COMMODITY
        elif m == 141:
            exchange = Exchange.ZCE
            typ = OptionsType.OPTIONS_COMMODITY
        elif m == 142:
            exchange = Exchange.INE
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 151:
            exchange = Exchange.SHFE
            typ = OptionsType.OPTIONS_COMMODITY
        elif m == 155:
            exchange = Exchange.LSE
            if t == 1:
                typ = StockType.STOCK_OTHER
            elif t == 2:
                typ = StockType.STOCK_P
            elif t == 3:
                typ = FundType.FUND_ETF
        elif m == 156:
            # s = [1, 5, 6, 7, 8]
            exchange = Exchange.LSE
            typ = DrType.DR_GDR
        elif m == 163:
            exchange = Exchange.INE
            typ = OptionsType.OPTIONS_COMMODITY
        elif m == 252:
            exchange = Exchange.SIX
            if t == 1:
                typ = DrType.DR_GDR
        elif m == 220:
            exchange = Exchange.CFFEX
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 221:
            exchange = Exchange.CFFEX
            typ = OptionsType.OPTIONS_INDEX
        elif m == 225:
            exchange = Exchange.GFEX
            typ = FuturesType.FUTURES_COMMODITY
        elif m == 226:
            exchange = Exchange.GFEX
            typ = OptionsType.OPTIONS_COMMODITY
        elif m == 341:
            exchange = Exchange.LSE
            typ = IndexType.INDEX_OTHER

        if typ and typ.parent() == SymbolType.FUTURES:
            if s & 1:
                # 主连合约
                flg.is_main_perpetual_contract = True  # 主力合约=True
            elif s & 2:
                # 次主连合约
                flg.is_second_main_perpetual_contract = True  # 次主力合约=True
            elif s & 4:
                flg.is_main_contract = True             # 主力合约
            elif s & 8:
                flg.is_second_main_contract = True      # 次主力合约

        return exchange, typ, flg

    def parse_str(self, fn) -> str | None:
        value: str | None = self.strip(fn=fn)
        if isinstance(value, str) and value and value != '-':
            return value.strip()

    def parse_int(self, fn: str) -> int | None:
        value: int | None = self.strip(fn=fn)
        if isinstance(value, int):
            return value

    def parse_float(self, fn: str) -> int | float | None:
        value: int | float | None = self.strip(fn=fn)
        if isinstance(value, (int, float)):
            return value

    def parse_yyyymmdd(self, fn: str) -> str | None:
        value: int | str | None = self.strip(fn=fn)
        if isinstance(value, int):
            value: str = str(value)
        if isinstance(value, str) and len(value) == 8:
            return datetime.datetime.strptime(value, '%Y%m%d').strftime(self.fmt_date)

    @property
    def m(self) -> int:
        return self.parse_int(fn='f13')

    @property
    def t(self) -> int:
        return self.parse_int(fn='f19')

    @property
    def s(self) -> int:
        return self.parse_int(fn='f111')

    @property
    def e(self) -> int:
        return self.parse_int(fn='f139')

    @property
    def f(self) -> int:
        return self.parse_int(fn='f148')

    @property
    def secid(self) -> str:
        return f'{self.m}.{self.code}'

    @property
    def trade_periods(self) -> Dict[str, List[List[str]]]:
        """解析交易时段"""
        def to_datetime(t: int) -> str:
            """时间转换YYYYmmddHHMM -> YYYY-mm-dd HH:MM:SS"""
            return datetime.datetime.strptime(str(t), '%Y%m%d%H%M').strftime('%Y-%m-%d %H:%M:%S')

        f400: Dict | None = self.strip(fn='f400')
        results: Dict[str, List[List[str]]] = {}
        if isinstance(f400, dict):
            for k, v in {'pre': 'O', 'after': 'C'}.items():
                p: Dict | None = f400.get(k, {})
                if p and isinstance(p, dict):
                    results[v] = [[to_datetime(p[k]) for k in ['b', 'e']]]
                else:
                    results[v] = []
            ps = f400.get('periods', [])
            if ps and isinstance(ps, list):
                results['T'] = [[to_datetime(p[k]) for k in ['b', 'e']] for p in ps]
            else:
                results['T'] = []
        return results

    @property
    def trade_date(self) -> str | None:
        """返回正在进行或即将开始的交易日"""
        f400: Dict | None = self.strip(fn='f400')
        dates: Set[str] = set()
        if isinstance(f400, dict):
            for k, v in f400.items():
                if isinstance(v, dict):
                    for be in ['b', 'e']:
                        if be in v:
                            dates.add(v[be])
                elif isinstance(v, list):
                    for _v in v:
                        if isinstance(_v, dict):
                            for be in ['b', 'e']:
                                if be in _v:
                                    dates.add(_v[be])
        if dates:
            return datetime.datetime.strptime(str(max(dates)), '%Y%m%d%H%M').strftime('%Y-%m-%d')

    @property
    def code(self):
        """标的代码"""
        return self.parse_str(fn='f12')

    @property
    def pre_close(self) -> int | float | None:
        """昨收价"""
        return self.parse_float(fn='f18')

    @property
    def total_shares(self) -> int | float | None:
        """总股本"""
        return self.parse_float(fn='f38')

    @property
    def float_shares(self) -> int | float | None:
        """流通股本"""
        return self.parse_float(fn='f39')

    @property
    def lot_size(self) -> int | float | None:
        """每手股数"""
        return self.parse_float(fn='f324')

    @property
    def apply_code(self) -> str | None:
        """申购代码"""
        return self.parse_str(fn='f348')

    @property
    def name_cn(self) -> str | None:
        """中文名称"""
        return self.parse_str(fn='f14')

    @property
    def name_en(self) -> str | None:
        """英文名称"""
        return self.parse_str(fn='f422')

    @property
    def underlying_code(self) -> str | None:
        """基础证券代码"""
        # 可转债的正股代码为f232
        # ETF期权的ETF代码为f331
        return self.parse_str(fn='f232') or self.parse_str(fn='f331')

    @property
    def py(self) -> str | None:
        """拼音简称"""
        name_cn: str = self.name_cn
        if name_cn:
            py = ''.join(
                pypinyin.lazy_pinyin(
                    re.sub(r'[\s,;.]', '', name_cn.translate(Trans)),
                    pypinyin.Style.FIRST_LETTER
                )
            ).lower()
            return py

    @property
    def list_date(self) -> str | None:
        """上市日期"""
        return self.parse_yyyymmdd(fn='f26')



# -*- coding: utf-8 -*-
"""国内交易所及部分国外交易所枚举"""
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Set
# https://www.iso20022.org/market-identifier-codes
# https://www.tradinghours.com/mic?country[]=CN&page=1#results


@dataclass(frozen=True)
class ExchangeDetail:
    acronym: str
    mic: str = ''
    extra_ids: List[str] = field(default_factory=list)   # 其他行情源的市场代码
    en_name: str = ''
    cn_name: str = ''
    website: str = ''

    @property
    def market_ids(self) -> Set[str]:
        extra_ids = {*self.extra_ids, self.acronym}
        if self.mic:
            extra_ids.add(self.mic)
        return extra_ids


class Exchange(Enum):
    """交易所简称枚举"""
    BSE = ExchangeDetail(acronym='BSE', mic='BJSE', extra_ids=['BJ', 'BJE', 'BJEX'], en_name='BEIJING STOCK EXCHANGE', cn_name='北京证券交易所', website='www.bse.cn')
    CFFEX = ExchangeDetail(acronym='CFFEX', mic='CCFX', extra_ids=['IF', 'CFX'], en_name='CHINA FINANCIAL FUTURES EXCHANGE', cn_name='中国金融期货交易所', website='www.cffex.com.cn')
    NEEQ = ExchangeDetail(acronym='NEEQ', mic='NEEQ', en_name='NATIONAL EQUITIES EXCHANGE AND QUOTATIONS', cn_name='全国中小企业股份转让系统', website='www.neeq.com.cn')
    SGE = ExchangeDetail(acronym='SGE', mic='SGEX', en_name='SHANGHAI GOLD EXCHANGE', cn_name='上海黄金交易所', website='www.sge.sh')
    CFETS = ExchangeDetail(acronym='CFETS', mic='XCFE', en_name='CHINA FOREIGN EXCHANGE TRADE SYSTEM', cn_name='中国外汇交易系统', website='www.chinamoney.com.cn')
    SSE = ExchangeDetail(acronym='SSE', mic='XSHG', extra_ids=['SH', 'SHO', 'SHOP'], en_name='SHANGHAI STOCK EXCHANGE', cn_name='上海证券交易所', website='www.szse.cn')
    SZSE = ExchangeDetail(acronym='SZSE', mic='XSHE', extra_ids=['SZ', 'SZO', 'SZOP'], en_name='SHENZHEN STOCK EXCHANGE', cn_name='深圳证券交易所', website='www.szse.cn')
    CSI = ExchangeDetail(acronym='CSI', mic='', en_name='CHINA SECURITIES INDEX', cn_name='中证指数', website='www.csindex.com.cn')
    SHFE = ExchangeDetail(acronym='SHFE', mic='XSGE', extra_ids=['SF', 'SHF'], en_name='SHANGHAI FUTURES EXCHANGE', cn_name='上海期货交易所', website='www.shfe.com.cn')
    INE = ExchangeDetail(acronym='INE', mic='XINE', en_name='SHANGHAI INTERNATIONAL ENERGY EXCHANGE', cn_name='上海国际能源交易中心', website='www.ine.cn')
    DCE = ExchangeDetail(acronym='DCE', mic='XDCE', extra_ids=['DF'], en_name='DALIAN COMMODITY EXCHANGE', cn_name='大连商品交易所', website='www.dce.com.cn')
    CZCE = ExchangeDetail(acronym='CZCE', mic='XZCE', extra_ids=['ZF', 'ZCE'], en_name='ZHENGZHOU COMMODITY EXCHANGE', cn_name='郑州商品交易所', website='www.czce.com.cn')
    GFEX = ExchangeDetail(acronym='GFEX', mic='', extra_ids=['GF', 'GFE'], en_name='GUANGZHOU COMMODITY EXCHANGE', cn_name='广州商品交易所', website='www.gfex.com.cn')
    CBOT = ExchangeDetail(acronym='CBOT', mic='CBTS', en_name='CME SWAPS MARKETS (CBOT)', cn_name='芝加哥期货交易所', website='www.cmegroup.com')
    COMEX = ExchangeDetail(acronym='COMEX', mic='CECS', en_name='CME SWAPS MARKETS (COMEX)', cn_name='纽约商品交易所', website='www.cmegroup.com')
    CME = ExchangeDetail(acronym='CME', mic='CMES', en_name='CME SWAPS MARKETS (CME)', cn_name='芝加哥商业交易所', website='www.cmegroup.com')
    NYMEX = ExchangeDetail(acronym='NYMEX', mic='NYMS', en_name='CME SWAPS MARKETS (NYMEX)', cn_name='纽约商品交易所', website='www.cmegroup.com')
    NYBOT = ExchangeDetail(acronym='NYBOT', mic='', en_name='THE NEW YORK BOARD OF TRADE', cn_name='纽约期货交易所', website='www.nybot.com')
    LME = ExchangeDetail(acronym='LME', mic='XLME', en_name='LONDON METAL EXCHANGE', cn_name='伦敦金属交易所', website='www.lme.co.uk')
    TOCOM = ExchangeDetail(acronym='TOCOM', mic='XTKT', en_name='TOKYO COMMODITY EXCHANGE', cn_name='东京商品交易所', website='www.tocom.or.jp')
    SICOM = ExchangeDetail(acronym='SICOM', mic='XSCE', en_name='SINGAPORE COMMODITY EXCHANGE', cn_name='新加坡商品交易所', website='www.sgx.com')
    HKEX = ExchangeDetail(acronym='HKEX', mic='XHKG', extra_ids=['HK', 'HGT', 'SGT'], en_name='HONG KONG EXCHANGES AND CLEARING LTD', cn_name='香港交易及结算所', website='www.hkex.com.hk')
    NASDAQ = ExchangeDetail(acronym='NASDAQ', mic='XNAS', en_name='NASDAQ - ALL MARKETS', cn_name='纳斯达克股票电子交易市场', website='www.nasdaq.com')
    NYSE = ExchangeDetail(acronym='NYSE', mic='XNYS', en_name='NEW YORK STOCK EXCHANGE, INC.', cn_name='纽约证券交易所', website='www.nyse.com')
    AMEX = ExchangeDetail(acronym='AMEX', mic='XASE', en_name='NYSE MKT LLC', cn_name='美国证券交易所', website='www.nyse.com')
    LSE = ExchangeDetail(acronym='LSE', mic='XLON', en_name='LONDON STOCK EXCHANGE', cn_name='伦敦证券交易所', website='www.lodonstockexchange.com')
    SIX = ExchangeDetail(acronym='SIX', mic='XSWX', en_name='SIX SWISS EXCHANGE', cn_name='瑞士证券交易所', website='www.six-group.com')
    ICE = ExchangeDetail(acronym='ICE', mic='', en_name='INTERCONTINENTAL EXCHANGE', cn_name='洲际交易所', website='www.theice.com')
    BMD = ExchangeDetail(acronym='BMD', mic='XKLS', en_name='BURSA MALAYSIA DERIVATIVES', cn_name='马来西亚衍生品交易所', website='www.bursamalaysia.com')
    SGX = ExchangeDetail(acronym='SGX', mic='XSES', en_name='SINGAPORE EXCHANGE', cn_name='新加坡交易所', website='www.sgx.com')

    @classmethod
    def find(cls, market_id: str) -> ExchangeDetail | None:
        """通过交易市场代码匹配返回交易所枚举"""
        return ExchangeMap.get(market_id)


ExchangeMap: Dict[str, Exchange] = {}

# 数据完整性和一致性检验
for e in Exchange:
    assert e.value.acronym == e.name
    for m in e.value.market_ids:
        ExchangeMap[m] = e



# -*- coding: utf-8 -*-
import brotli
import humanize
import io
import lzma
import zlib
import zstandard
from enum import Enum
from loguru import logger


class Compressor(Enum):
    IDENTITY = 'identity'
    DEFLATE = 'deflate'
    GZIP = 'gzip'
    ZLIB = 'zlib'
    BR = 'br'
    ZSTD = 'zstd'

    def get_compressor(self, level: int = 1):
        """获取流式压缩对象"""
        if self == self.DEFLATE:
            return zlib.compressobj(level=level, method=zlib.DEFLATED, wbits=-zlib.MAX_WBITS)
        elif self == self.GZIP:
            return zlib.compressobj(level=level, method=zlib.DEFLATED, wbits=zlib.MAX_WBITS | 16)
        elif self == self.ZLIB:
            return zlib.compressobj(level=level, method=zlib.DEFLATED, wbits=zlib.MAX_WBITS)
        elif self == self.BR:
            return brotli.Compressor()
        elif self == self.ZSTD:
            return zstandard.ZstdCompressor(level)


class ZstdStreamCompressor:
    def __init__(self, size: int = -1, interval: int = -1, threads: int = 0, level: int = zstandard.STRATEGY_BTULTRA2):
        """
        with ZstdStreamCompressor(size=len(data1) + len(data2), interval=1024*1024) as sc:
            sc.update(data=data1)
            sc.update(data=data2)
            return sc.stream
        """
        self.stream: io.BytesIO = io.BytesIO()
        self.size: int = size
        self.interval: int = interval
        self.interval_count: int = 0
        self.compressor = zstandard.ZstdCompressor(level=level, threads=threads)
        self.chunker = self.compressor.chunker(size=self.size)
        self._origin_size: int = 0

    def update(self, data: bytes):
        """对原始数据流进行流式压缩"""
        assert isinstance(data, bytes)
        buffer = io.BytesIO(data)
        while block := buffer.read(16777216):
            self._origin_size += len(block)
            for compressed in self.chunker.compress(block):
                self.stream.write(compressed)
                self.log()
        return self.stream

    def log(self, force: bool = False):
        if self.interval >= 0 and (force or self.stream.tell() >= self.interval_count * self.interval):
            prefix: str = f'{self._origin_size / self.size:6.1%}' if self.size > 0 else ''
            logger.debug(f'[{prefix}]已压缩={humanize.naturalsize(self.stream.tell())} 压缩比={self.ratio:6.1%}')
            self.interval_count += 1

    @property
    def origin_size(self) -> int:
        """压缩前字节数"""
        return self._origin_size

    @property
    def compressed_size(self) -> int:
        """压缩后字节数"""
        return self.stream.tell()

    @property
    def ratio(self) -> float:
        """压缩率"""
        origin_size: int = self.origin_size
        size: int = self.compressed_size
        if origin_size > 0:
            return size / origin_size
        else:
            return 1.0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None and exc_val is None and exc_tb is None:
            try:
                for compressed in self.chunker.flush():
                    self.stream.write(compressed)
                    self.log()
                for compressed in self.chunker.finish():
                    self.stream.write(compressed)
                    self.log(force=True)
            except Exception as e:
                self.stream.close()
                raise e
        else:
            self.stream.close()


class LzmaStreamCompressor:
    def __init__(self):
        self.stream: io.BytesIO = io.BytesIO()
        self.compressor = lzma.LZMACompressor(preset=lzma.PRESET_EXTREME)

    def update(self, data: bytes):
        """对原始数据流进行流式压缩"""
        assert isinstance(data, bytes)
        buffer = io.BytesIO(data)
        while block := buffer.read(16777216):
            compressed: bytes = self.compressor.compress(block)
            self.stream.write(compressed)
        return self.stream

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None and exc_val is None and exc_tb is None:
            try:
                self.stream.write(self.compressor.flush())
            except Exception as e:
                self.stream.close()
                raise e
        else:
            self.stream.close()



# -*- coding: utf-8 -*-
"""国内银行枚举"""
from collections import namedtuple
from enum import auto, Enum
from typing import List, Dict
from .base import ByKey

# https://www.swift.com/

field_names: List[str] = ["swift_code", "cn_name"]
BankDetail = namedtuple(typename="BankDetail", field_names=field_names)
tsv: str = """
BKCHCNBJ\t中国银行
ABOCCNBJ\t农业银行
PCBCCNBJ\t建设银行
COMMCNSH\t交通银行
CHBHCNBT\t渤海银行
CIBKCNBJ\t中信银行
CMBCCNBS\t招商银行
EVERCNBJ\t光大银行
MSBCCNBJ\t民生银行
HXBKCNBJ\t华夏银行
GDBKCN22\t广发银行
FJIBCNBA\t兴业银行
SPDBCNSH\t上海浦东发展银行
SDBCCNBJ\t国家开发银行
SZDBCNBS\t平安银行
HFBACNSD\t恒丰银行
HSBCCNSH\t汇丰银行（中国）有限公司
BEASCNSH\t东亚银行（中国）有限公司
SCBLCNSX\t渣打银行（中国）有限公司
HZCBCN2H\t杭州银行
HCCBCNBH\t哈尔滨银行
HASECNSH\t恒生银行
BJCNCNBJ\t北京银行
BKNBCN2N\t宁波银行
JZCBCNBJ\t锦州银行
CITICNSX\t花旗银行（中国）有限公司
DBSSCNSH\t星展银行（中国）有限公司
BKSHCNBJ\t河北银行
IBXHCNBA\t厦门国际银行
CBXMCNBA\t厦门银行
ZBBKCNBZ\t齐商银行
HSSYCNBH\t内蒙古银行
BOSHCNSH\t上海银行
WZCBCNSH\t温州银行
BOTKCNSH\t三菱东京日联银行（中国）有限公司
RZCBCNBD\t日照银行
BTCBCNBJ\t包头银行
IBKOCNBT\t韩国兴业银行（中国）有限公司
MHCBCNSH\t瑞穗银行（中国）有限公司
DEUTCNSH\t德意志银行（中国）有限公司
LWCBCNBJ\t莱商银行
ANZBCNSH\t澳大利亚及新西兰银行（中国）有限公司
SHBKCNBJ\t新韩银行（中国）有限公司
CZNBCNBJ\t国民银行（中国）有限公司
BKKBCNSH\t曼谷银行（中国）有限公司
CCBCCNBK\t长华商业银行
CKLBCNBJ\t昆仑银行
DYSHCNBJ\t东营银行
CIYUCNBA\t赤渝银行
FSBCCNSH\t富邦银行（中国）有限公司
ICBCCNBS\t大型国际商业银行
WPACCNSX\t西太平洋银行
QCCBCNBQ\t青岛银行
DECLCNBJ\t德州银行
FCBKCNSH\t第一商业银行
DSBACNBX\t大新银行（中国）有限公司
MSBKCN22\t摩根士丹利国际银行(中国)有限公司
TACBCNBS\t台湾合作银行
PSBCCNBJ\t中国邮政储蓄银行
BINHCN2N\t宁波商业银行
CHASCNBJ\t摩根大通银行（中国）有限公司
LCHBCN22\t创兴银行
UWCBCNSH\t国泰联合银行（中国）有限公司
YCCBCNBY\t宁夏银行
OUBKCNB1\t海外联合银行
LYCBCNBL\t临商银行
BOJJCNBJ\t九江银行
CHCCCNSS\t长沙银行
BOJSCNBN\t江苏银行
HNBKCNBS\t华南商业银行
KCCBCN2K\t富滇银行
SCBKCNBS\t上海商业银行
CRLYCNSH\t法国农业信贷银行（中国）有限公司
WFCBCNBN\t潍坊银行
RZBACNBJ\t瑞丰国际银行
UOVBCNSH\t联合海外银行（中国）有限公司
SBINCNSH\t印度国家银行
NJCBCNBN\t南京银行
BOFMCNBJ\t蒙特利尔银行（中国）有限公司
DLCBCNBD\t大连银行
BNPACNSH\t法国巴黎银行（中国）有限公司
SYCBCNBY\t盛京银行
TCCBCNBT\t天津银行
HRXJCNBC\t华容银行
MBTCCNBN\t大都会银行（中国）有限公司
RBOSCNS1\t苏格兰皇家银行
APGTCN21\t亚太银行
CTGBCN22\t重庆三峡银行
CCIBCNB1\t中国信贷产业银行
BOFMCNSH\t蒙特利尔银行（中国）有限公司
GDPBCN22\t广东南岳银行
SSVBCNSH\t浦发硅谷银行
ZCCBCN22\t珠海华润银行
WHCBCNBN\t汉口银行
WHBKCNBJ\t威海市商业银行
HVBKCNBJ\t友利银行（中国）有限公司
HFCBCNSH\t徽商银行
FXBKCNBJ\t阜新银行
TZBKCNBT\t泰州银行
YZBKCN2N\t鄞州银行
KWHKCNBS\t中信银行国际（中国）有限公司
SWEDCNSH\t瑞典银行
FZCBCNBS\t福建海下银行
BKJNCNBJ\t济宁银行
CQCBCN22\t重庆银行
COXICNBA\t联合商业银行
DEUTCNBJ\t德意志银行（中国）有限公司
YTCBCNSD\t烟台银行
ZJMTCNSH\t浙江民泰商业银行
NXBKCNBH\t南浔银行
EWBKCNSH\t华美银行（中国）有限公司
JHCBCNBJ\t金华银行
BOFMCN22\t蒙特利尔银行（中国）有限公司
LCOMCNBJ\t辽阳银行
GDHBCN22\t广东华兴银行
KRTHCNB1\t泰国国立银行
COBACNSX\t德国商业银行
YKCBCNBJ\t营口银行
SCBLCNB1\t渣打银行
SMBCCNSH\t三井住友银行（中国）有限公司
ZJCBCN2N\t浙商银行
CTBACNSH\t澳大利亚联邦银行
KREDCNSX\t比利时联合银行
LJBCCNBH\t龙江银行
BKQZCNBZ\t泉州银行
CBOCCNBC\t成都银行
CNMBCNBS\t中国商业银行
BSHLCNS1\t巴林银行
PASCCNSH\t锡耶纳蒙特银行
KWPBCNB1\t广东省银行
GZCBCN22\t广州银行
LSCCCNBL\t乐山市商业银行
GYCBCNSI\t贵阳银行
"""
BankDetails: List[BankDetail] = [BankDetail(*line.strip().split("\t")) for line in tsv.splitlines() if line]


class BySwiftCode(ByKey):
    map: Dict[str, BankDetail] = {getattr(item, field_names[0]): item for item in BankDetails}


class Bank(Enum):
    BKCHCNBJ = BySwiftCode.map['BKCHCNBJ']
    ABOCCNBJ = BySwiftCode.map['ABOCCNBJ']
    PCBCCNBJ = BySwiftCode.map['PCBCCNBJ']
    COMMCNSH = BySwiftCode.map['COMMCNSH']
    CHBHCNBT = BySwiftCode.map['CHBHCNBT']
    CIBKCNBJ = BySwiftCode.map['CIBKCNBJ']
    CMBCCNBS = BySwiftCode.map['CMBCCNBS']
    EVERCNBJ = BySwiftCode.map['EVERCNBJ']
    MSBCCNBJ = BySwiftCode.map['MSBCCNBJ']
    HXBKCNBJ = BySwiftCode.map['HXBKCNBJ']
    GDBKCN22 = BySwiftCode.map['GDBKCN22']
    FJIBCNBA = BySwiftCode.map['FJIBCNBA']
    SPDBCNSH = BySwiftCode.map['SPDBCNSH']
    SDBCCNBJ = BySwiftCode.map['SDBCCNBJ']
    SZDBCNBS = BySwiftCode.map['SZDBCNBS']
    HFBACNSD = BySwiftCode.map['HFBACNSD']
    HSBCCNSH = BySwiftCode.map['HSBCCNSH']
    BEASCNSH = BySwiftCode.map['BEASCNSH']
    SCBLCNSX = BySwiftCode.map['SCBLCNSX']
    HZCBCN2H = BySwiftCode.map['HZCBCN2H']
    HCCBCNBH = BySwiftCode.map['HCCBCNBH']
    HASECNSH = BySwiftCode.map['HASECNSH']
    BJCNCNBJ = BySwiftCode.map['BJCNCNBJ']
    BKNBCN2N = BySwiftCode.map['BKNBCN2N']
    JZCBCNBJ = BySwiftCode.map['JZCBCNBJ']
    CITICNSX = BySwiftCode.map['CITICNSX']
    DBSSCNSH = BySwiftCode.map['DBSSCNSH']
    BKSHCNBJ = BySwiftCode.map['BKSHCNBJ']
    IBXHCNBA = BySwiftCode.map['IBXHCNBA']
    CBXMCNBA = BySwiftCode.map['CBXMCNBA']
    ZBBKCNBZ = BySwiftCode.map['ZBBKCNBZ']
    HSSYCNBH = BySwiftCode.map['HSSYCNBH']
    BOSHCNSH = BySwiftCode.map['BOSHCNSH']
    WZCBCNSH = BySwiftCode.map['WZCBCNSH']
    BOTKCNSH = BySwiftCode.map['BOTKCNSH']
    RZCBCNBD = BySwiftCode.map['RZCBCNBD']
    BTCBCNBJ = BySwiftCode.map['BTCBCNBJ']
    IBKOCNBT = BySwiftCode.map['IBKOCNBT']
    MHCBCNSH = BySwiftCode.map['MHCBCNSH']
    DEUTCNSH = BySwiftCode.map['DEUTCNSH']
    LWCBCNBJ = BySwiftCode.map['LWCBCNBJ']
    ANZBCNSH = BySwiftCode.map['ANZBCNSH']
    SHBKCNBJ = BySwiftCode.map['SHBKCNBJ']
    CZNBCNBJ = BySwiftCode.map['CZNBCNBJ']
    BKKBCNSH = BySwiftCode.map['BKKBCNSH']
    CCBCCNBK = BySwiftCode.map['CCBCCNBK']
    CKLBCNBJ = BySwiftCode.map['CKLBCNBJ']
    DYSHCNBJ = BySwiftCode.map['DYSHCNBJ']
    CIYUCNBA = BySwiftCode.map['CIYUCNBA']
    FSBCCNSH = BySwiftCode.map['FSBCCNSH']
    ICBCCNBS = BySwiftCode.map['ICBCCNBS']
    WPACCNSX = BySwiftCode.map['WPACCNSX']
    QCCBCNBQ = BySwiftCode.map['QCCBCNBQ']
    DECLCNBJ = BySwiftCode.map['DECLCNBJ']
    FCBKCNSH = BySwiftCode.map['FCBKCNSH']
    DSBACNBX = BySwiftCode.map['DSBACNBX']
    MSBKCN22 = BySwiftCode.map['MSBKCN22']
    TACBCNBS = BySwiftCode.map['TACBCNBS']
    PSBCCNBJ = BySwiftCode.map['PSBCCNBJ']
    BINHCN2N = BySwiftCode.map['BINHCN2N']
    CHASCNBJ = BySwiftCode.map['CHASCNBJ']
    LCHBCN22 = BySwiftCode.map['LCHBCN22']
    UWCBCNSH = BySwiftCode.map['UWCBCNSH']
    YCCBCNBY = BySwiftCode.map['YCCBCNBY']
    OUBKCNB1 = BySwiftCode.map['OUBKCNB1']
    LYCBCNBL = BySwiftCode.map['LYCBCNBL']
    BOJJCNBJ = BySwiftCode.map['BOJJCNBJ']
    CHCCCNSS = BySwiftCode.map['CHCCCNSS']
    BOJSCNBN = BySwiftCode.map['BOJSCNBN']
    HNBKCNBS = BySwiftCode.map['HNBKCNBS']
    KCCBCN2K = BySwiftCode.map['KCCBCN2K']
    SCBKCNBS = BySwiftCode.map['SCBKCNBS']
    CRLYCNSH = BySwiftCode.map['CRLYCNSH']
    WFCBCNBN = BySwiftCode.map['WFCBCNBN']
    RZBACNBJ = BySwiftCode.map['RZBACNBJ']
    UOVBCNSH = BySwiftCode.map['UOVBCNSH']
    SBINCNSH = BySwiftCode.map['SBINCNSH']
    NJCBCNBN = BySwiftCode.map['NJCBCNBN']
    BOFMCNBJ = BySwiftCode.map['BOFMCNBJ']
    DLCBCNBD = BySwiftCode.map['DLCBCNBD']
    BNPACNSH = BySwiftCode.map['BNPACNSH']
    SYCBCNBY = BySwiftCode.map['SYCBCNBY']
    TCCBCNBT = BySwiftCode.map['TCCBCNBT']
    HRXJCNBC = BySwiftCode.map['HRXJCNBC']
    MBTCCNBN = BySwiftCode.map['MBTCCNBN']
    RBOSCNS1 = BySwiftCode.map['RBOSCNS1']
    APGTCN21 = BySwiftCode.map['APGTCN21']
    CTGBCN22 = BySwiftCode.map['CTGBCN22']
    CCIBCNB1 = BySwiftCode.map['CCIBCNB1']
    BOFMCNSH = BySwiftCode.map['BOFMCNSH']
    GDPBCN22 = BySwiftCode.map['GDPBCN22']
    SSVBCNSH = BySwiftCode.map['SSVBCNSH']
    ZCCBCN22 = BySwiftCode.map['ZCCBCN22']
    WHCBCNBN = BySwiftCode.map['WHCBCNBN']
    WHBKCNBJ = BySwiftCode.map['WHBKCNBJ']
    HVBKCNBJ = BySwiftCode.map['HVBKCNBJ']
    HFCBCNSH = BySwiftCode.map['HFCBCNSH']
    FXBKCNBJ = BySwiftCode.map['FXBKCNBJ']
    TZBKCNBT = BySwiftCode.map['TZBKCNBT']
    YZBKCN2N = BySwiftCode.map['YZBKCN2N']
    KWHKCNBS = BySwiftCode.map['KWHKCNBS']
    SWEDCNSH = BySwiftCode.map['SWEDCNSH']
    FZCBCNBS = BySwiftCode.map['FZCBCNBS']
    BKJNCNBJ = BySwiftCode.map['BKJNCNBJ']
    CQCBCN22 = BySwiftCode.map['CQCBCN22']
    COXICNBA = BySwiftCode.map['COXICNBA']
    DEUTCNBJ = BySwiftCode.map['DEUTCNBJ']
    YTCBCNSD = BySwiftCode.map['YTCBCNSD']
    ZJMTCNSH = BySwiftCode.map['ZJMTCNSH']
    NXBKCNBH = BySwiftCode.map['NXBKCNBH']
    EWBKCNSH = BySwiftCode.map['EWBKCNSH']
    JHCBCNBJ = BySwiftCode.map['JHCBCNBJ']
    BOFMCN22 = BySwiftCode.map['BOFMCN22']
    LCOMCNBJ = BySwiftCode.map['LCOMCNBJ']
    GDHBCN22 = BySwiftCode.map['GDHBCN22']
    KRTHCNB1 = BySwiftCode.map['KRTHCNB1']
    COBACNSX = BySwiftCode.map['COBACNSX']
    YKCBCNBJ = BySwiftCode.map['YKCBCNBJ']
    SCBLCNB1 = BySwiftCode.map['SCBLCNB1']
    SMBCCNSH = BySwiftCode.map['SMBCCNSH']
    ZJCBCN2N = BySwiftCode.map['ZJCBCN2N']
    CTBACNSH = BySwiftCode.map['CTBACNSH']
    KREDCNSX = BySwiftCode.map['KREDCNSX']
    LJBCCNBH = BySwiftCode.map['LJBCCNBH']
    BKQZCNBZ = BySwiftCode.map['BKQZCNBZ']
    CBOCCNBC = BySwiftCode.map['CBOCCNBC']
    CNMBCNBS = BySwiftCode.map['CNMBCNBS']
    BSHLCNS1 = BySwiftCode.map['BSHLCNS1']
    PASCCNSH = BySwiftCode.map['PASCCNSH']
    KWPBCNB1 = BySwiftCode.map['KWPBCNB1']
    GZCBCN22 = BySwiftCode.map['GZCBCN22']
    LSCCCNBL = BySwiftCode.map['LSCCCNBL']
    GYCBCNSI = BySwiftCode.map['GYCBCNSI']


# 数据完整性和一致性检验
BySwiftCode.check(obj=Bank)
